from typing import Any, Dict, List
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages import BaseMessage
from langchain_core.outputs import LLMResult
from langchain_core.agents import AgentAction, AgentFinish
import json
from llumo.llumoLogger import LlumoLogger
from llumo.llumoSessionContext import LlumoSessionContext
import time
import re


class LlumoCallbackHandler(BaseCallbackHandler):
    def __init__(self, session: LlumoSessionContext = None,agentType = "react_agent"):
        if session is None:
            raise ValueError("LlumoSessionContext is required")

        self.sessionLogger = session
        self.agentType = agentType

        # Initialize timing and state variables
        self.llmStartTime = None
        self.agentStartTime = None
        self.toolStartTime = None
        self.chainStartTime = None
        self.stepTime = None

        # Initialize tracking variables
        self.prompt = ""
        self.searchQuery = ""
        self.currentToolName = None
        self.currentToolInput = None
        self.currentAgentName = None
        self.currentChainTime = "unknown"
        self.agentsSteps = 0
        self.toolsUsed = []
        self.llmProvider = "unknown"

        # Status tracking
        self.hasErrors = False
        self.lastError = None

        # ReAct specific tracking
        self.reactSteps = []
        self.currentThought = ""
        self.currentAction = ""
        self.currentObservation = ""
        self.isAgentExecution = False



    def on_chain_start(self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any) -> None:
        """Called when a chain starts - this includes agent execution"""
        # print("ON CHAIN START: ",inputs)
        # print("ON CHAIN START: serialized",serialized)
        # print("ON CHAIN START: kwargs",kwargs)

        try:
            self.prompt = inputs.get("input", "")
            self.chainStartTime = time.time()

            # Reset all tracking variables for new query
            self.agentsSteps = 0
            self.toolsUsed = []
            self.reactSteps = []
            self.currentThought = ""
            self.currentAction = ""
            self.currentObservation = ""
            self.currentToolName = None
            self.currentToolInput = None
            self.hasErrors = False
            self.lastError = None
            self.toolDescription = {}

            # Dynamically detect agent name from serialized data
            if serialized is not None:
                # Check for 'name' parameter passed during initialization
                if "name" in serialized:
                    self.currentAgentName = serialized["name"]
                elif "_type" in serialized:
                    self.currentAgentName = serialized["_type"]
                elif "kwargs" in serialized and "name" in serialized["kwargs"]:
                    self.currentAgentName = serialized["kwargs"]["name"]
                elif "id" in serialized and isinstance(serialized["id"], list):
                    self.currentAgentName = serialized["id"][-1] if serialized["id"] else "unknown"
                else:
                    self.currentAgentName = "unknown"

            # Check if this is agent execution
            if ("agent" in str(self.currentAgentName).lower() or
                    (serialized and serialized.get("_type") == "agent_executor") or
                    any(key in str(serialized).lower() for key in ["agent", "react", "executor"])):

                self.agentStartTime = time.time()
                self.isAgentExecution = True
                print(f"[DEBUG] Agent execution started: {self.currentAgentName} - Reset counters for new query")
            else:
                self.isAgentExecution = False

        except Exception as e:
            print(f"[ERROR] in on_chain_start: {e}")

        try:
            self.sessionLogger.logQueryStep(
                stepName = "Query Invocation",
                model = "unknown",
                provider = "unknown",
                inputTokens = round(len(self.prompt.split()) * 1.5),
                query = self.prompt,
                status = "SUCCESS"
            )
        except Exception as e:
            self.sessionLogger.logQueryStep(
                stepName="Query Invocation",
                model="unknown",
                provider="unknown",
                inputTokens=0,
                query="",
                status="FAILURE"
            )
            print(f"[ERROR] Failed to log user input: {e}")

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Called when a chain ends"""
        try:
            if self.isAgentExecution and isinstance(outputs, dict) and "output" in outputs:
                # Use logAgentStep for final completion
                self.sessionLogger.logAgentStep(
                    stepName="Agent Execution Completed",
                    agentType=self.agentType,
                    agentName=self.currentAgentName or "unknown",
                    numStepsTaken=self.agentsSteps,
                    tools=self.toolsUsed,
                    query=self.prompt,
                    status="SUCCESS",
                    # message=f"Final output: {outputs['output']}. ReAct steps: {json.dumps(self.reactSteps)}",
                )


            # Reset execution state after chain ends
            self.isAgentExecution = False

        except Exception as e:
            print(f"[ERROR] Failed to log chain output: {e}")

    def on_llm_start(self, serialized: Dict[str, Any], prompts: List[Any], **kwargs: Any) -> None:
        """Called when LLM starts"""
        self.llmStartTime = time.time()
        self.stepTime = time.time()

        if self.prompt == "":
            match = re.search(r"Human:\s*(.*)",prompts[0], re.DOTALL)
            if match:
                user_question = match.group(1).strip()
                self.prompt = user_question # 👉 What is LangChain?
            else:
                self.prompt = ""

        # Dynamically get model info
        model = "unknown"
        if serialized and "kwargs" in serialized:
            model = serialized["kwargs"].get("model_name",serialized["kwargs"].get("model", "unknown"))

        provider = "unknown"
        if isinstance(serialized.get("id", []), list) and len(serialized["id"]) > 2:
            provider = serialized["id"][2]
        self.llmProvider = provider

    def on_llm_end(self, response: Any, **kwargs: Any) -> None:
        """Called when LLM completes"""
        duration_ms = int((time.time() - self.llmStartTime) * 1000) if self.llmStartTime else 0

        # Initialize default values
        output = ""
        model_name = "unknown"
        input_tokens = 0
        output_tokens = 0
        status = "SUCCESS"
        error_message = ""

        try:
            # Case 1: LLMResult object
            if hasattr(response, 'generations'):
                if response.generations and len(response.generations) > 0:
                    generation = response.generations[0][0] if len(response.generations[0]) > 0 else None
                    if generation:
                        # Handle message content
                        if hasattr(generation, 'message'):
                            message = generation.message
                            if hasattr(message, 'content'):
                                output = message.content
                            # Handle function calls
                            if hasattr(message, 'additional_kwargs'):
                                func_call = message.additional_kwargs.get('function_call')
                                if func_call:
                                    output = (f"Function call: {func_call.get('name', 'unknown')} "
                                            f"with arguments {func_call.get('arguments', '{}')}")
                            if not output:
                                output = str(message)
                        else:
                            output = getattr(generation, 'text', str(generation))

                # Get token usage and model name
                if hasattr(response, 'llm_output'):
                    model_name = response.llm_output.get("model_name", "unknown")
                    usage = response.llm_output.get("token_usage", {})
                    input_tokens = usage.get("prompt_tokens", usage.get("input_tokens", 0))
                    output_tokens = usage.get("completion_tokens", usage.get("output_tokens", 0))

            # Case 2: Direct AIMessage
            elif hasattr(response, 'content'):
                output = response.content
                # Handle function calls
                if hasattr(response, 'additional_kwargs'):
                    func_call = response.additional_kwargs.get('function_call')
                    if func_call:
                        output = (f"Function call: {func_call.get('name', 'unknown')} "
                                    f"with arguments {func_call.get('arguments', '{}')}")

                # Get metadata
                if hasattr(response, 'response_metadata'):
                    model_name = getattr(response.response_metadata, "model_name", "unknown")
                    token_usage = getattr(response.response_metadata, "token_usage", {})
                    if isinstance(token_usage, dict):
                        input_tokens = token_usage.get("prompt_tokens", token_usage.get("input_tokens", 0))
                        output_tokens = token_usage.get("completion_tokens", token_usage.get("output_tokens", 0))

            # Case 3: Other types
            else:
                output = str(response)
                if hasattr(response, 'model'):
                    model_name = response.model

        except Exception as e:
            error_message = f"Response processing error: {str(e)}"
            status = "ERROR"

        # Ensure we have string values
        output = str(output) if output is not None else ""
        model_name = str(model_name) if model_name is not None else "unknown"

        # Parse ReAct reasoning from LLM output if we're in agent execution
        if self.isAgentExecution and output:
            self._parse_react_reasoning(output)
        try:
            self.sessionLogger.logLlmStep(
                stepName="LLM Call Completed",
                model=model_name,
                provider=self.llmProvider,
                inputTokens=int(input_tokens),
                outputTokens=int(output_tokens),
                # temperature=float(kwargs.get("temperature", 0.7)),
                # promptTruncated=False,
                latencyMs=duration_ms,
                prompt=str(self.prompt),
                output=output,
                status=status,
                # message=error_message if status == "ERROR" else "",
            )

        except Exception as e:
            print(f"[ERROR] Failed to log LLM end: {e}")

    def on_retriever_start(self, serialized, query, run_id, parent_run_id=None, **kwargs):
        self.prompt = query
        self.searchQuery = query

    def on_retriever_end(self, documents, run_id, parent_run_id=None, **kwargs):

        try:
            chunkSize = len(documents[0].page_content) if documents and documents[0].page_content else 0
        except Exception:
            chunkSize = 0

        source = ( kwargs.get("metadata", {}).get("source") or kwargs.get("tags") or "unknown")

        try:
            self.sessionLogger.logRetrieverStep(
                stepName="Context Retrieval Complete",
                retrieverSource = str(source),
                topK = len(documents),
                chunkSize =  chunkSize,
                context =  [doc.page_content for doc in documents],
                searchQuery =  self.prompt if self.prompt != "" else self.searchQuery,
                latencyMs = 120,  # mock latency, replace with real timing if needed
                status = "SUCCESS"
             )
        except Exception as e:
            print(f"[ERROR] Failed to log chain output: {e}")

    def on_retriever_error(self, error, run_id, parent_run_id=None, **kwargs):

        try:
            self.sessionLogger.logRetrieverStep(
                stepName="Context Retrieval Error",
                retrieverSource = kwargs.get("metadata", {}).get("source", "unknown"),
                topK = 0,
                chunkSize =  0,
                context = [],
                searchQuery =  self.prompt if self.prompt != "" else self.searchQuery,
                latencyMs = 0,  # mock latency, replace with real timing if needed
                status = "FAILURE"
             )
        except Exception as e:
            print(f"[ERROR] Failed to log chain output: {e}")


    def _parse_react_reasoning(self, llm_output: str):
        """Parse ReAct reasoning pattern from LLM output"""
        try:
            # Extract thought patterns
            thought_match = re.search(r'Thought:\s*(.+?)(?=Action:|$)', llm_output, re.DOTALL)
            if thought_match:
                self.currentThought = thought_match.group(1).strip()

            # Extract action patterns
            action_match = re.search(r'Action:\s*(.+?)(?=Action Input:|Thought:|$)', llm_output, re.DOTALL)
            if action_match:
                self.currentAction = action_match.group(1).strip()

            # Extract action input patterns
            action_input_match = re.search(r'Action Input:\s*(.+?)(?=Observation:|Thought:|$)', llm_output, re.DOTALL)
            action_input = ""
            if action_input_match:
                action_input = action_input_match.group(1).strip()

            # Store the reasoning step for ReAct trace
            if self.currentThought or self.currentAction:
                reasoning_step = {
                    "step_number": self.agentsSteps + 1,
                    "thought": self.currentThought,
                    "planned_action": self.currentAction,
                    "action_input": action_input,
                    "full_llm_output": llm_output,
                    "timestamp": time.time()
                }

                # Add to react steps for complete trace
                if not self.reactSteps or self.reactSteps[-1].get("step_number") != reasoning_step["step_number"]:
                    self.reactSteps.append(reasoning_step)
                else:
                    # Update existing step
                    self.reactSteps[-1].update(reasoning_step)

        except Exception as e:
            print(f"[ERROR] Failed to parse ReAct reasoning: {e}")

    def on_tool_start(self, serialized: Dict[str, Any], input_str: str, **kwargs: Any) -> None:
        """Called when a tool starts executing"""
        # print("ON TOOL START: ",serialized)
        # print("ON TOOL START: ",kwargs)

        self.toolStartTime = time.time()
        self.stepTime = time.time()

        # Dynamically get tool name
        self.currentToolName = (serialized.get("name") or
                                serialized.get("_type") or
                                "unknown")
        self.currentToolDescription = serialized.get("description","No description found")

        # Handle the case where input_str is "None" or None
        if input_str == "None" or input_str is None:
            self.currentToolInput = {"input": ""}
        else:
            try:
                # Try to parse as JSON if it looks like JSON
                if input_str.startswith("{") and input_str.endswith("}"):
                    self.currentToolInput = json.loads(input_str)
                else:
                    self.currentToolInput = {"input": input_str}
            except:
                self.currentToolInput = {"input": input_str}

        # Track tools used
        if self.currentToolName not in self.toolsUsed:
            self.toolsUsed.append(self.currentToolName)

        print(f"[DEBUG] Tool started: {self.currentToolName} with input: {input_str}")

    def on_tool_end(self, output: Any, **kwargs: Any) -> None:
        """Called when a tool completes execution"""
        duration_ms = int((time.time() - self.toolStartTime) * 1000) if self.toolStartTime else 0

        try:
            # Ensure output is stringified safely
            if output is None:
                output_str = ""
            elif isinstance(output, (dict, list)):
                output_str = json.dumps(output)
            else:
                output_str = str(output)

            # Store as observation for ReAct step
            self.currentObservation = output_str

            # Update the current ReAct step with observation
            if self.reactSteps and self.isAgentExecution:
                self.reactSteps[-1]["observation"] = output_str
                self.reactSteps[-1]["tool_execution_ms"] = duration_ms

            self.sessionLogger.logToolStep(
                stepName="Tool Execution Completed",
                toolName=self.currentToolName or "unknown",
                description = self.currentToolDescription,
                input=self.currentToolInput or {"input": ""},
                output=output_str,
                latencyMs=duration_ms,
                status="SUCCESS",
                # message="",
            )
            print(f"[DEBUG] Tool completed: {self.currentToolName} -> {output_str}")

        except Exception as e:
            print(f"[ERROR] Failed to log tool end: {e}")

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> None:
        """Called when an agent takes an action"""
        self.agentsSteps += 1
        # print("ON AGENT ACTION: ", action)

        try:
            # Dynamically extract information from action
            tool_name = getattr(action, "tool", "unknown")
            tool_input = getattr(action, "tool_input", "")
            log_message = getattr(action, "log", "")

            # Track tools if not already tracked
            if tool_name not in self.toolsUsed:
                self.toolsUsed.append(tool_name)

            # Update our ReAct steps tracking with executed action
            if self.reactSteps:
                self.reactSteps[-1].update({
                    "executed_action": tool_name,
                    "executed_input": tool_input,
                    "action_log": log_message
                })

            # Log the agent action step using logAgentStep
            current_status = "FAILURE" if self.hasErrors else "SUCCESS"
            reasoning_text = self.currentThought if self.currentThought else "No reasoning captured"

            self.sessionLogger.logAgentStep(
                stepName=f"Agent Action Step {self.agentsSteps}",
                agentType=self.agentType,
                agentName=self.currentAgentName or "unknown",
                numStepsTaken=self.agentsSteps,
                tools=[tool_name],
                query=self.prompt,
                status=current_status,
                # message=f"Executing {tool_name} with input: {tool_input}. Reasoning: {reasoning_text}",
            )



        except Exception as e:
            print(f"[ERROR] Failed to log agent action: {e}")

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Called when an agent completes execution"""
        # print("ON AGENT FINISH:", finish)
        # We don't need to log anything here since the final result is already logged in on_chain_end
        pass

    def on_agent_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when an agent encounters an error"""
        # print("ITS A AGENT ERROR:", error)
        self.hasErrors = True
        self.lastError = str(error)

        try:
            self.sessionLogger.logAgentStep(
                stepName="Agent Execution Error",
                agentType=self.agentType,
                agentName=self.currentAgentName or "unknown",
                numStepsTaken=self.agentsSteps,
                tools=self.toolsUsed,
                query=self.prompt,
                status="FAILURE",
                # message=str(error),
            )
        except Exception as e:
            print(f"[ERROR] Failed to log agent error: {e}")

    def on_tool_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when a tool encounters an error"""

        self.hasErrors = True
        self.lastError = str(error)

        try:
            # Update ReAct step with error observation
            if self.reactSteps:
                self.reactSteps[-1]["observation"] = f"ERROR: {str(error)}"
                self.reactSteps[-1]["error"] = True

            self.sessionLogger.logToolStep(
                stepName="Tool Execution Failed",
                toolName=self.currentToolName or "unknown",
                description=self.currentToolDescription,
                input=self.currentToolInput or {"input": ""},
                output="",
                latencyMs=0,
                status="FAILURE",
                # message=str(error),
            )
        except Exception as e:
            print(f"[ERROR] Failed to log tool error: {e}")

    def on_chain_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when a chain encounters an error"""
        # print("ITS A CHAIN ERROR:", error)
        self.hasErrors = True
        self.lastError = str(error)

        try:
            if self.isAgentExecution:
                # Use logAgentStep for agent-related chain errors
                self.sessionLogger.logAgentStep(
                    stepName="Agent Chain Error",
                    agentType=self.agentType,
                    agentName=self.currentAgentName or "unknown",
                    numStepsTaken=self.agentsSteps,
                    tools=self.toolsUsed,
                    query=self.prompt,
                    status="FAILURE",
                    # message=str(error),
                )



            else:
                # Use logLlmStep for general chain errors
                self.sessionLogger.logLlmStep(
                    stepName="Chain Execution Error",
                    model="unknown",
                    provider=self.llmProvider,
                    inputTokens=0,
                    outputTokens=0,
                    # temperature=0.0,
                    # promptTruncated=False,
                    latencyMs=0,
                    prompt=self.prompt,
                    output="",
                    status="FAILURE",
                    # message=str(error),
                )
        except Exception as e:
            print(f"[ERROR] Failed to log chain error: {e}")

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Called when LLM generates a new token (for streaming)"""
        pass

    def on_text(self, text: str, **kwargs: Any) -> None:
        """Called when arbitrary text is logged"""
        # Only log significant text events during agent execution
        if self.isAgentExecution and text.strip():
            print(f"[DEBUG] Additional text: {text}")

            # Check if this text contains important ReAct information like "Observation:"
            if any(keyword in text.lower() for keyword in ['observation:']):
                try:
                    # Update the current ReAct step with additional observation info
                    if self.reactSteps:
                        existing_obs = self.reactSteps[-1].get("observation", "")
                        self.reactSteps[-1][
                            "observation"] = f"{existing_obs}\n{text.strip()}" if existing_obs else text.strip()

                except Exception as e:
                    print(f"[ERROR] Failed to process text event: {e}")



import requests


class LlumoLogger:
    def __init__(self, apiKey: str, playground: str):
        self.apiKey = apiKey
        self.playground = playground
        self.workspaceID = None
        self.playgroundID = None
        self.userEmailID = None
        self._authenticate()

    def _authenticate(self):
        url = "https://app.llumo.ai/api/get-playground-name"
        try:
            response = requests.post(
                url,
                headers={
                    "Authorization": f"Bearer {self.apiKey}",
                    "Content-Type": "application/json",
                },
                json={"playgroundName": self.playground},
                timeout=10,
            )

            response.raise_for_status()
            res_json = response.json()

            # Navigate into the nested "data" structure
            inner_data = res_json.get("data", {}).get("data", {})

            self.workspaceID = inner_data.get("workspaceID")
            self.playgroundID = inner_data.get("playgroundID")
            self.userEmailID = inner_data.get("createdBy")

            if not self.workspaceID or not self.playgroundID:
                raise RuntimeError(
                    f"Invalid response: workspaceID or playgroundID missing. Full response: {res_json}"
                )

        except requests.exceptions.RequestException as req_err:
            raise RuntimeError(
                f"Network or HTTP error during authentication: {req_err}"
            )
        except ValueError as json_err:
            raise RuntimeError(f"Invalid JSON in authentication response: {json_err}")
        except Exception as e:
            raise RuntimeError(f"Authentication failed: {e}")

    def getWorkspaceID(self):
        return self.workspaceID

    def getUserEmailID(self):
        return self.userEmailID

    def getPlaygroundID(self):
        return self.playgroundID



from google import generativeai as _genai
from .client import LlumoClient


def evaluate_multiple(data, api_key=None, evals=["Response Correctness"]):
    client = LlumoClient(api_key=api_key)
    results = client.evaluateMultiple(
        data,
        evals=evals,
        createExperiment=False,
        prompt_template="Give answer to the query: {{query}}, using context: {{context}}",
        getDataFrame=False
    )
    return results


class ChatCompletionWithEval:
    def __init__(self, response, evaluation=None):
        self._response = response
        # self.evaluation = evaluation

    def __getattr__(self, name):
        return getattr(self._response, name)

    def __getitem__(self, key):
        return self._response[key]

    def __repr__(self):
        return repr(self._response)


class genai:
    class GenerativeModel:
        def __init__(self, api_key: str, model: str = "gemini-2.5-flash"):
            _genai.configure(api_key=api_key)
            self._api_key = api_key
            self._model_name = model
            self._model_instance = _genai.GenerativeModel(model_name=model)

        def generate_content(self, contents: str | list[str], **kwargs):
            context = kwargs.pop("context", None)
            evals = kwargs.pop("evals", [])
            llumo_key = kwargs.pop("llumo_key", None)

            # Run Gemini generation
            response = self._model_instance.generate_content(contents=contents, **kwargs)
            output = response.text

            # eval_input = [{
            #     "query": contents,
            #     "context": context or contents,
            #     "output": output,
            # }]

            # evaluation = None
            # try:
            #     evaluation = evaluate_multiple(data=eval_input, evals=evals, api_key=llumo_key)
            # except Exception as e:
            #     evaluation = None
            
            # if evaluation is None:
            #         print("Cannot process your request for evaluation, please check your api and try again later.")
            #         return response
            

            return ChatCompletionWithEval(response, evaluation=None)



from enum import Enum

class Provider(str, Enum):
    OPENAI = "OPENAI"
    GOOGLE = "GOOGLE"

# Maps model aliases → (provider, actual model name for API)
_MODEL_METADATA = {
    "GPT_4O": (Provider.OPENAI, "GPT_4O"),
    "GPT_4_5": (Provider.OPENAI, "GPT_4_5"),
    "GPT_4": (Provider.OPENAI, "GPT_4"),
    "GPT_4_32K": (Provider.OPENAI, "GPT_4_32K"),
    "GPT_3_5_Turbo": (Provider.OPENAI, "GPT_35T"),
    "GPT_3_5_Turbo_Instruct": (Provider.OPENAI, "GPT_35T_INS"),
    "GPT_3_5_Turbo_16K": (Provider.OPENAI, "GPT_35T_16K"),
    "GPT_4_o_Mini": (Provider.OPENAI, "GPT_4O_MINI"),
    "o4_MINI": (Provider.OPENAI, "O4_MINI"),
    "o4_MINI_HIGH": (Provider.OPENAI, "O4_MINI_HIGH"),
    "GPT_4_1": (Provider.OPENAI, "GPT_4_1"),
    "GPT_4_1_Mini": (Provider.OPENAI, "GPT_4_1_MINI"),
    "GPT_4_1_nano": (Provider.OPENAI, "GPT_4_1_NANO"),
    "o3": (Provider.OPENAI, "O3"),
    "o3_MINI": (Provider.OPENAI, "O3_MINI"),
    "o1": (Provider.OPENAI, "O1"),
    "o1_MINI": (Provider.OPENAI, "O1_MINI"),
    
    
    "Gemini_2_5_Pro": (Provider.GOOGLE, "GEMINI_2_5_PRO"),
    "Gemini_2_5_Flash": (Provider.GOOGLE, "GEMINI_2_5_FLASH"),
    "Gemini_2_0": (Provider.GOOGLE, "GEMINI_2_0"),
    "Gemini_2_0_Flash": (Provider.GOOGLE, "GEMINI_2_0_FLASH"),
    "Gemini_Pro": (Provider.GOOGLE, "GEMINI_PRO"),
    "Text_Bison": (Provider.GOOGLE, "TEXT_BISON"),
    "Chat_Bison": (Provider.GOOGLE, "CHAT_BISON"),
    "Text_Bison_32k": (Provider.GOOGLE, "TEXT_BISON_32K"),
    "Text_Unicorn": (Provider.GOOGLE, "TEXT_UNICORN"),
    "Google_1_5_Flash": (Provider.GOOGLE, "GOOGLE_15_FLASH"),
    "Gemma_3_9B": (Provider.GOOGLE, "GEMMA_3_9B"),
    "Gemma_3_27B": (Provider.GOOGLE, "GEMMA_3_27B"),
}

class AVAILABLEMODELS(str, Enum):
    GPT_4o= "GPT_4O",
    GPT_4o_Mini= "GPT_4O_MINI",
    GPT_4_5= "GPT_4_5",
    GPT_4= "GPT_4",
    GPT_4_32K= "GPT_4_32K",
    GPT_3_5_Turbo= "GPT_35T",
    GPT_3_5_Turbo_Instruct= "GPT_35T_INS",
    GPT_3_5_Turbo_16K= "GPT_35T_16K",
    GPT_4_o_Mini= "GPT_4O_MINI",
    o4_MINI = "O4_MINI",
    o4_MINI_HIGH = "O4_MINI_HIGH",
    GPT_4_1 = "GPT_4_1",
    GPT_4_1_Mini = "GPT_4_1_MINI",
    GPT_4_1_nano = "GPT_4_1_NANO",
    o3 = "O3",
    o3_MINI = "O3_MINI",
    o1 = "O1",
    o1_MINI = "O1_MINI",
    
    Gemini_2_5_Pro = "GEMINI_2_5_PRO",
    Gemini_2_5_Flash = "GEMINI_2_5_FLASH",
    Gemini_2_0 = "GEMINI_2_0",
    Gemini_2_0_Flash = "GEMINI_2_0_FLASH",
    Gemini_Pro = "GEMINI_PRO",
    Text_Bison = "TEXT_BISON",
    Chat_Bison = "CHAT_BISON",
    Text_Bison_32k = "TEXT_BISON_32K",
    Text_Unicorn = "TEXT_UNICORN",
    Google_1_5_Flash = "GOOGLE_15_FLASH",
    Gemma_3_9B = "GEMMA_3_9B",
    Gemma_3_27B = "GEMMA_3_27B",
    

def getProviderFromModel(model: AVAILABLEMODELS) -> Provider:
    for alias, (provider, apiName) in _MODEL_METADATA.items():
        if model.value == apiName:
            return provider
    raise ValueError(f"Provider not found for model: {model}")


import requests


import time
import re
import json
import uuid
import warnings
import os
import itertools
import pandas as pd
from typing import List, Dict
from .models import AVAILABLEMODELS, getProviderFromModel, Provider
from .execution import ModelExecutor
from .exceptions import LlumoAIError
from .helpingFuntions import *
from .sockets import LlumoSocketClient
from .functionCalling import LlumoAgentExecutor
from .chains import LlumoDataFrameResults, LlumoDictResults
import threading
from tqdm import tqdm

pd.set_option("future.no_silent_downcasting", True)

postUrl = "https://redskull.llumo.ai/api/process-playground"

# postUrl = "http://localhost:4747/api/process-playground"
fetchUrl = "https://redskull.llumo.ai/api/get-cells-data"

# fetchUrl = "http://localhost:4747/api/get-cells-data"
fetchMissingEvalUrl = "https://redskull.llumo.ai/api/get-missing-keys"
socketDataUrl = "https://redskull.llumo.ai/api/eval/get-awaited"

validateUrl = "https://app.llumo.ai/api/workspace-details"
socketUrl = "https://redskull.llumo.ai/"
# socketUrl = "http://localhost:4747/"
createEvalUrl = "https://backend-api.llumo.ai/api/v1/create-debug-log-for-sdk"
# createEvalUrl = "http://localhost:4545/api/v1/create-debug-log-for-sdk"


class LlumoClient:

    def __init__(self, api_key, playground_id=None):
        self.apiKey = api_key
        self.playgroundID = playground_id
        self.evalData = []
        self.evals = []
        self.processMapping = {}
        self.definationMapping = {}

    def validateApiKey(self, evalName="Input Bias"):
        headers = {
            "Authorization": f"Bearer {self.apiKey}",
            "Content-Type": "application/json",
        }
        reqBody = {"analytics": [evalName]}

        try:

            response = requests.post(url=validateUrl, json=reqBody, headers=headers)

        except requests.exceptions.RequestException as e:
            print(f"Request exception: {str(e)}")
            raise LlumoAIError.RequestFailed(detail=str(e))

        if response.status_code == 401:
            raise LlumoAIError.InvalidApiKey()

        # Handle other common status codes
        if response.status_code == 404:
            raise LlumoAIError.RequestFailed(
                detail=f"Endpoint not found (404): {validateUrl}"
            )

        if response.status_code != 200:
            raise LlumoAIError.RequestFailed(
                detail=f"Unexpected status code: {response.status_code}"
            )

        # Try to parse JSON
        try:
            data = response.json()
            # print(data)
        except ValueError as e:
            print(f"JSON parsing error: {str(e)}")
            # print(f"Response content that could not be parsed: {response.text[:1000]}...")
            raise LlumoAIError.InvalidJsonResponse()

        if "data" not in data or not data["data"]:
            # print(f"Invalid API response structure: {data}")
            raise LlumoAIError.InvalidApiResponse()

        try:
            self.hitsAvailable = data["data"]["data"].get("remainingHits", 0)
            self.workspaceID = data["data"]["data"].get("workspaceID")
            self.evalDefinition = data["data"]["data"]["analyticsMapping"]
            self.socketToken = data["data"]["data"].get("token")
            # print(self.socketToken)
            self.hasSubscribed = data["data"]["data"].get("hasSubscr"
                                                          "ibed", False)
            self.trialEndDate = data["data"]["data"].get("trialEndDate", None)
            self.subscriptionEndDate = data["data"]["data"].get(
                "subscriptionEndDate", None
            )
            self.email = data["data"]["data"].get("email", None)

            self.definationMapping[evalName] = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
            )
            self.categories = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("categories", {})
            )
            self.evaluationStrictness = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("evaluationStrictness", {})
            )
            self.grammarCheckOutput = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("grammarCheckOutput", {})
            )
            self.insightsLength = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("insightsLength", {})
            )
            self.insightsLevel = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("insightsLevel", {})
            )
            self.executionDependency = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("executionDependency", {})
            )
            self.sampleData = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("sampleData", {})
            )
            self.numJudges = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("numJudges", {})
            )
            self.penaltyBonusInstructions = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("penaltyBonusInstructions", [])
            )
            self.probableEdgeCases = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("probableEdgeCases", [])
            )
            self.fieldMapping = (
                data.get("data", {})
                .get("data", {})
                .get("analyticsMapping", {})
                .get(evalName, "")
                .get("fieldMapping", [])
            )

        except Exception as e:
            # print(f"Error extracting data from response: {str(e)}")
            raise LlumoAIError.UnexpectedError(detail=evalName)

    def postBatch(self, batch, workspaceID):
        payload = {
            "batch": json.dumps(batch),
            "runType": "FULL_EVAL_RUN",
            "workspaceID": workspaceID,
        }
        # socketToken here if the "JWD" token
        headers = {
            "Authorization": f"Bearer {self.socketToken}",
            "Content-Type": "application/json",
        }
        try:
            print(postUrl)
            response = requests.post(postUrl, json=payload, headers=headers)
            print(f"Post API Status Code: {response.status_code}")
            print(response.text)
            # print(response.status_code)

        except Exception as e:
            print(f"Error in posting batch: {e}")

    
    def fetchDataForMissingKeys(self, workspaceID, missingKeys: list):
    # Define the URL and prepare the payload
       
        payload = {
            "workspaceID": workspaceID,
            "missingKeys": missingKeys
        }

        headers = {
            "Authorization": f"Bearer {self.socketToken}",
            "Content-Type": "application/json",
        }
        
        try:
            # Send a POST request to the API
            response = requests.post(fetchMissingEvalUrl, json=payload, headers=headers)
            
            # Check if the response is successful
            if response.status_code == 200:
                # Parse the JSON data from the response
                data = response.json().get("data", {})
                
                
                # Prepare the list of all data values in the desired format
                result_list = []
                for key, value in data.items():
                    # Create a dictionary for each item in the response data
                    result_list.append({
                        key: {
                            "value": value.get("value"),
                            "fullEval": value.get("fullEval", {}),
                            "runLog": value.get("runLog", {}),
                            "evallist": value.get("evallist", [])
                        }
                    })
                
                print("Fetched data for missing keys:", result_list)
                return result_list
            else:
                print(f"Failed to fetch data. Status Code: {response.status_code}")
                return []

        except Exception as e:
            print(f"An error occurred: {e}")
            return []

    
    def createRunForEvalMultiple(self, smartLog):
        payload = {
            "log": smartLog
        }
        # socketToken here if the "JWD" token
        headers = {
            # "Authorization": f"Bearer {self.socketToken}",
            "Content-Type": "application/json",
        }
        try:
            # print(postUrl)
            print(createEvalUrl)
            print(payload)
            print(headers)
            response = requests.post(createEvalUrl, json=payload, headers=headers)

            print(f"Post API Status Code: {response.status_code}")
            print(response.text)
            # print(response.status_code)

        except Exception as e:
            print(f"Error in posting batch: {e}")


    def postDataStream(self, batch, workspaceID):
        payload = {
            "batch": json.dumps(batch),
            "runType": "DATA_STREAM",
            "workspaceID": workspaceID,
        }
        # socketToken here if the "JWD" token
        headers = {
            "Authorization": f"Bearer {self.socketToken}",
            "Content-Type": "application/json",
        }
        try:
            # print(postUrl)
            response = requests.post(postUrl, json=payload, headers=headers)
            # print(f"Post API Status Code: {response.status_code}")
            # print(response.text)

        except Exception as e:
            print(f"Error in posting batch: {e}")

    def AllProcessMapping(self):
        for batch in self.allBatches:
            for record in batch:
                rowId = record["rowID"]
                colId = record["columnID"]
                pid = f"{rowId}-{colId}-{colId}"
                self.processMapping[pid] = record

    def finalResp(self, results):
        seen = set()
        uniqueResults = []

        for item in results:
            for rowID in item:  # Each item has only one key
                # for rowID in item["data"]:
                if rowID not in seen:
                    seen.add(rowID)
                    uniqueResults.append(item)

        return uniqueResults

    # this function allows the users to run exactl one eval at a time
    def evaluate(
        self,
        data,
        eval="Response Completeness",
        prompt_template="",
        outputColName="output",
        createExperiment: bool = False,
        _tocheck=True,
    ):

        # converting it into a pandas dataframe object
        dataframe = pd.DataFrame(data)

        # check for dependencies for the selected eval metric
        metricDependencies = checkDependency(
            eval, columns=list(dataframe.columns), tocheck=_tocheck
        )
        if metricDependencies["status"] == False:
            raise LlumoAIError.dependencyError(metricDependencies["message"])

        results = {}
        try:
            socketID = self.socket.connect(timeout=150)

            # Ensure full connection before proceeding
            max_wait_secs = 20
            waited_secs = 0
            while not self.socket._connection_established.is_set():
                time.sleep(0.1)
                waited_secs += 0.1
                if waited_secs >= max_wait_secs:
                    raise RuntimeError(
                        "Timeout waiting for server 'connection-established' event."
                    )

            rowIdMapping = {}

            print(f"\n======= Running evaluation for: {eval} =======")

            try:
                self.validateApiKey(evalName=eval)
            except Exception as e:
                if hasattr(e, "response") and getattr(e, "response", None) is not None:
                    pass
                raise
            userHits = checkUserHits(
                self.workspaceID,
                self.hasSubscribed,
                self.trialEndDate,
                self.subscriptionEndDate,
                self.hitsAvailable,
                len(dataframe),
            )

            if not userHits["success"]:
                raise LlumoAIError.InsufficientCredits(userHits["message"])

            # if self.hitsAvailable == 0 or len(dataframe) > self.hitsAvailable:
            #     raise LlumoAIError.InsufficientCredits()

            evalDefinition = self.evalDefinition[eval].get("definition")
            model = "GPT_4"
            provider = "OPENAI"
            evalType = "LLM"
            workspaceID = self.workspaceID
            email = self.email

            self.allBatches = []
            currentBatch = []

            for index, row in dataframe.iterrows():
                tools = [row["tools"]] if "tools" in dataframe.columns else []
                groundTruth = (
                    row["groundTruth"] if "groundTruth" in dataframe.columns else ""
                )
                messageHistory = (
                    [row["messageHistory"]]
                    if "messageHistory" in dataframe.columns
                    else []
                )
                promptTemplate = prompt_template

                keys = re.findall(r"{{(.*?)}}", promptTemplate)

                if not all([ky in dataframe.columns for ky in keys]):
                    raise LlumoAIError.InvalidPromptTemplate()

                inputDict = {key: row[key] for key in keys if key in row}
                output = (
                    row[outputColName] if outputColName in dataframe.columns else ""
                )

                activePlayground = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace(
                    "-", ""
                )
                rowID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
                columnID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")

                # storing the generated rowID and the row index (dataframe) for later lookkup
                rowIdMapping[rowID] = index

                templateData = {
                    "processID": getProcessID(),
                    "socketID": socketID,
                    "source": "SDK",
                    "processData": {
                        "executionDependency": {
                            "query": "",
                            "context": "",
                            "output": output,
                            "tools": tools,
                            "groundTruth": groundTruth,
                            "messageHistory": messageHistory,
                        },
                        "definition": evalDefinition,
                        "model": model,
                        "provider": provider,
                        "analytics": eval,
                    },
                    "workspaceID": workspaceID,
                    "type": "EVAL",
                    "evalType": evalType,
                    "kpi": eval,
                    "columnID": columnID,
                    "rowID": rowID,
                    "playgroundID": activePlayground,
                    "processType": "EVAL",
                    "email": email,
                }

                query = ""
                context = ""
                for key, value in inputDict.items():
                    if isinstance(value, str):
                        length = len(value.split()) * 1.5
                        if length > 50:
                            context += f" {key}: {value}, "
                        else:
                            if promptTemplate:
                                tempObj = {key: value}
                                promptTemplate = getInputPopulatedPrompt(
                                    promptTemplate, tempObj
                                )
                            else:
                                query += f" {key}: {value}, "

                if not context.strip():
                    for key, value in inputDict.items():
                        context += f" {key}: {value}, "

                templateData["processData"]["executionDependency"][
                    "context"
                ] = context.strip()
                templateData["processData"]["executionDependency"][
                    "query"
                ] = query.strip()

                if promptTemplate and not query.strip():
                    templateData["processData"]["executionDependency"][
                        "query"
                    ] = promptTemplate

                currentBatch.append(templateData)

                if len(currentBatch) == 10 or index == len(dataframe) - 1:
                    self.allBatches.append(currentBatch)
                    currentBatch = []

            totalItems = sum(len(batch) for batch in self.allBatches)

            for cnt, batch in enumerate(self.allBatches):
                try:

                    self.postBatch(batch=batch, workspaceID=workspaceID)
                    print("Betch Posted with item len: ", len(batch))
                except Exception as e:
                    continue

                # time.sleep(3)

            timeout = max(50, min(600, totalItems * 10))

            self.socket.listenForResults(
                min_wait=40,
                max_wait=timeout,
                inactivity_timeout=150,
                expected_results=totalItems,
            )

            eval_results = self.socket.getReceivedData()
            results[eval] = self.finalResp(eval_results)

        except Exception as e:
            raise
        finally:
            try:
                self.socket.disconnect()
            except Exception as e:
                pass

        for evalName, records in results.items():
            dataframe[evalName] = None
            for item in records:
                for compound_key, value in item.items():
                    # for compound_key, value in item['data'].items():

                    rowID = compound_key.split("-")[0]
                    # looking for the index of each rowID , in the original dataframe
                    if rowID in rowIdMapping:
                        index = rowIdMapping[rowID]
                        # dataframe.at[index, evalName] = value
                        dataframe.at[index, evalName] = value["value"]
                        dataframe.at[index, f"{evalName} Reason"] = value["reasoning"]

                    else:
                        pass
                        # print(f"⚠️ Warning: Could not find rowID {rowID} in mapping")
        if createExperiment:
            pd.set_option("future.no_silent_downcasting", True)
            df = dataframe.fillna("Some error occured").astype(object)

            if createPlayground(
                email,
                workspaceID,
                df,
                promptText=prompt_template,
                definationMapping=self.definationMapping,
                outputColName=outputColName,
            ):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results.Please rerun the experiment to see the results on playground."
                )
        else:
            return dataframe

    # this function allows the users to run multiple evals at once

    def compressor(self, data, prompt_template):
        results = []
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        dataframe = pd.DataFrame(data)

        try:
            self.socket = LlumoSocketClient(socketUrl)
            dataframe = pd.DataFrame(data).astype(str)
            socketID = self.socket.connect(timeout=250)

            # Wait for socket connection
            max_wait_secs = 20
            waited_secs = 0
            while not self.socket._connection_established.is_set():
                time.sleep(0.1)
                waited_secs += 0.1
                if waited_secs >= max_wait_secs:
                    raise RuntimeError("Timeout waiting for server connection")

            # Start listener thread
            expectedResults = len(dataframe)
            # print("expected result" ,expectedResults)
            timeout = max(100, min(150, expectedResults * 10))
            listener_thread = threading.Thread(
                target=self.socket.listenForResults,
                kwargs={
                    "min_wait": 40,
                    "max_wait": timeout,
                    "inactivity_timeout": 10,
                    "expected_results": expectedResults,
                },
                daemon=True,
            )
            listener_thread.start()

            try:
                self.validateApiKey()
            except Exception as e:
                print(f"Error during API key validation: {str(e)}")
                if hasattr(e, "response") and getattr(e, "response", None) is not None:
                    print(f"Status code: {e.response.status_code}")
                    print(f"Response content: {e.response.text[:500]}...")
                raise

            userHits = checkUserHits(
                self.workspaceID,
                self.hasSubscribed,
                self.trialEndDate,
                self.subscriptionEndDate,
                self.hitsAvailable,
                len(dataframe),
            )

            if not userHits["success"]:
                raise LlumoAIError.InsufficientCredits(userHits["message"])

            model = "GPT_4"
            provider = "OPENAI"
            evalType = "LLUMO"
            workspaceID = self.workspaceID
            email = self.email
            self.allBatches = []
            currentBatch = []
            rowIdMapping = {}
            for index, row in dataframe.iterrows():
                promptTemplate = prompt_template
                keys = re.findall(r"{{(.*?)}}", promptTemplate)
                inputDict = {key: row[key] for key in keys if key in row}

                if not all([ky in dataframe.columns for ky in keys]):
                    raise LlumoAIError.InvalidPromptTemplate()

                activePlayground = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace(
                    "-", ""
                )
                rowID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
                columnID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")

                compressed_prompt_id = (
                    f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
                )
                compressed_prompt_output_id = (
                    f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
                )
                cost_id = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
                cost_saving_id = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace(
                    "-", ""
                )

                rowDataDict = {}
                for col in dataframe.columns:
                    val = row[col]
                    rowDataDict[col] = {"type": "VARIABLE", "value": str(val)}

                templateData = {
                    "processID": getProcessID(),
                    "socketID": socketID,
                    "source": "SDK",
                    "rowID": rowID,
                    "columnID": columnID,
                    "processType": "COST_SAVING",
                    "evalType": evalType,
                    "dependency": list(inputDict.keys()),
                    "costColumnMapping": {
                        "compressed_prompt": compressed_prompt_id,
                        "compressed_prompt_output": compressed_prompt_output_id,
                        "cost": cost_id,
                        "cost_saving": cost_saving_id,
                    },
                    "processData": {
                        "rowData": rowDataDict,
                        "dependency": list(inputDict.keys()),
                        "dependencyMapping": {ky: ky for ky in list(inputDict.keys())},
                        "provider": provider,
                        "model": model,
                        "promptText": promptTemplate,
                        "costColumnMapping": {
                            "compressed_prompt": compressed_prompt_id,
                            "compressed_prompt_output": compressed_prompt_output_id,
                            "cost": cost_id,
                            "cost_saving": cost_saving_id,
                        },
                    },
                    "workspaceID": workspaceID,
                    "email": email,
                    "playgroundID": activePlayground,
                }

                rowIdMapping[f"{rowID}-{columnID}-{columnID}"] = index
                # print("__________________________TEMPLATE__________________________________")
                # print(templateData)

                currentBatch.append(templateData)

                if len(currentBatch) == 10 or index == len(dataframe) - 1:
                    self.allBatches.append(currentBatch)
                    currentBatch = []

            total_items = sum(len(batch) for batch in self.allBatches)

            for cnt, batch in enumerate(self.allBatches):
                try:
                    self.postBatch(batch=batch, workspaceID=workspaceID)
                except Exception as e:
                    print(f"Error posting batch {cnt + 1}: {str(e)}")
                    continue
                time.sleep(1)

            self.AllProcessMapping()
            timeout = max(60, min(600, total_items * 10))
            self.socket.listenForResults(
                min_wait=20,
                max_wait=timeout,
                inactivity_timeout=50,
                expected_results=None,
            )

            rawResults = self.socket.getReceivedData()
            receivedRowIDs = {key for item in rawResults for key in item.keys()}
            expectedRowIDs = set(rowIdMapping.keys())
            missingRowIDs = expectedRowIDs - receivedRowIDs
            # print("All expected keys:", expected_rowIDs)
            # print("All received keys:", received_rowIDs)
            # print("Missing keys:", len(missingRowIDs))
            missingRowIDs = list(missingRowIDs)

            if len(missingRowIDs) > 0:
                dataFromDb = fetchData(workspaceID, activePlayground, missingRowIDs)
                rawResults.extend(dataFromDb)

            # results = self.finalResp(eval_results)
            # print(f"======= Completed evaluation: {eval} =======\n")

        except Exception as e:
            print(f"Error during evaluation: {e}")
            raise
        finally:
            try:
                self.socket.disconnect()
            except Exception as e:
                print(f"Error disconnecting socket: {e}")

        dataframe["Compressed Input"] = None
        for records in rawResults:
            for compound_key, value in records.items():
                # for compound_key, value in item['data'].items():
                rowID = compound_key
                # looking for the index of each rowID , in the original dataframe
                if rowID in rowIdMapping:
                    index = rowIdMapping[rowID]

                    dataframe.at[index, "Compressed Input"] = value["value"]

                else:
                    pass
                    # print(f"⚠️ Warning: Could not find rowID {rowID} in mapping")

        # compressed_prompt, compressed_prompt_output, cost, cost_saving = costColumnMapping(results, self.processMapping)
        # dataframe["compressed_prompt"] = compressed_prompt
        # dataframe["compressed_prompt_output"] = compressed_prompt_output
        # dataframe["cost"] = cost
        # dataframe["cost_saving"] = cost_saving

        return dataframe

    def evaluateMultiple(
        self,
        data,
        evals: list = [],
        # prompt_template="Give answer to the given query: {{query}} using the given context: {{context}}.",
        prompt_template="",
        getDataFrame: bool = False,
        _tocheck=True,
    ):
        # if hasattr(self, "startLlumoRun"):
        #     self.startLlumoRun(runName="evaluateMultiple")
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        self.socket = LlumoSocketClient(socketUrl)
        dataframe = pd.DataFrame(data).astype(str)
        workspaceID = None
        email = None
        try:
            socketID = self.socket.connect(timeout=250)
            print("Socket connected with ID:", socketID)
        except Exception as e:
            socketID = "DummySocketID"
            print(f"Socket connection failed, using dummy ID. Error: {str(e)}")

        self.evalData = []
        self.evals = evals
        self.allBatches = []
        rowIdMapping = {}  # (rowID-columnID-columnID -> (index, evalName))

        # Wait for socket connection
        # max_wait_secs = 20
        # waited_secs = 0
        # while not self.socket._connection_established.is_set():
        #     time.sleep(0.1)
        #     waited_secs += 0.1
        #     if waited_secs >= max_wait_secs:
        #         raise RuntimeError("Timeout waiting for server connection")

        # Start listener thread
        # expectedResults = len(dataframe) * len(evals)
        expectedResults = len(dataframe)
        # print("expected result" ,expectedResults)
        timeout = max(100, min(250, expectedResults * 60))
        listener_thread = threading.Thread(
            target=self.socket.listenForResults,
            kwargs={
                "min_wait": 20,
                "max_wait": timeout,
                "inactivity_timeout": timeout,
                "expected_results": expectedResults,
            },
            daemon=True,
        )
        listener_thread.start()
        # commenting validate api key as we don't need it logger does it for us. uncommented but we need different
        # api for this which don't spend time on eval defintiion fetches and just bring hits
        self.validateApiKey()
        activePlayground = self.playgroundID
        # print(f"\n======= Running evaluation for: {evalName} =======")

        # Validate API and dependencies
        # self.validateApiKey(evalName=evals[0])

        # why we need custom analytics here? there is no such usage below
        # customAnalytics = getCustomAnalytics(self.workspaceID)

        # metricDependencies = checkDependency(
        #     evalName,
        #     list(dataframe.columns),
        #     tocheck=_tocheck,
        #     customevals=customAnalytics,
        # )
        # if not metricDependencies["status"]:
        #     raise LlumoAIError.dependencyError(metricDependencies["message"])

        # evalDefinition = self.evalDefinition[evalName]["definition"]
        model = "GPT_4"
        provider = "OPENAI"
        evalType = "LLM"
        workspaceID = self.workspaceID
        email = self.email
        # categories = self.categories
        # evaluationStrictness = self.evaluationStrictness
        # grammarCheckOutput = self.grammarCheckOutput
        # insightLength = self.insightsLength
        # numJudges = self.numJudges
        # penaltyBonusInstructions = self.penaltyBonusInstructions
        # probableEdgeCases = self.probableEdgeCases
        # fieldMapping = self.fieldMapping

        userHits = checkUserHits(
            self.workspaceID,
            self.hasSubscribed,
            self.trialEndDate,
            self.subscriptionEndDate,
            self.hitsAvailable,
            len(dataframe),
        )

        #where does this remaining hit comes from?

        
        if not userHits["success"]:
            raise LlumoAIError.InsufficientCredits(userHits["message"])

        currentBatch = []
        

        for index, row in dataframe.iterrows():
            # Extract required fields
            tools = row.get("tools", "")
            groundTruth = row.get("groundTruth", "")
            messageHistory = row.get("messageHistory", "")
            intermediateSteps = row.get("intermediateSteps", "")
            output = row.get("output", "")
            
            # Initialize query and context
            query = ""
            context = ""
            
            # Process prompt template if provided
            if prompt_template:
                # Extract template variables
                keys = re.findall(r"{{(.*?)}}", prompt_template)
                
                if not all([key in dataframe.columns for key in keys]):
                    raise LlumoAIError.InvalidPromptTemplate()
                
                # Populate template and separate query/context
                populated_template = prompt_template
                for key in keys:
                    value = row.get(key, "")
                    if isinstance(value, str):
                        length = len(value.split()) * 1.5
                        if length <= 50:
                            # Short value - include in query via template
                            temp_obj = {key: value}
                            populated_template = getInputPopulatedPrompt(populated_template, temp_obj)
                        else:
                            # Long value - add to context
                            context += f" {key}: {value}, "
                
                query = populated_template.strip()
                
                # Add any remaining context from other fields
                if not context.strip():
                    for key, value in row.items():
                        if key not in keys and isinstance(value, str) and value.strip():
                            context += f" {key}: {value}, "
            else:
                # No prompt template - use direct query and context fields
                query = row.get("query", "")
                context = row.get("context", "")
            
            # Generate unique IDs
            rowID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
            columnID = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
            
            compoundKey = f"{rowID}-{columnID}-{columnID}"
            rowIdMapping[compoundKey] = {"index": index}
            print("rowIdMapping:", rowIdMapping)

            # Create evaluation payload
            print("socketID in before templateData: ", socketID)
            templateData = {
                "processID": getProcessID(),
                "socketID": socketID,
                "rowID": rowID,
                "columnID": columnID,
                "processType": "FULL_EVAL_RUN",
                "evalType": "LLM",
                "workspaceID": workspaceID,
                "email": email,
                "playgroundID": activePlayground,
                "source": "SDK",
                "processData": {
                    "executionDependency": {
                        "query": query,
                        "context": context.strip(),
                        "output": output,
                        "tools": tools,
                        "groundTruth": groundTruth,
                        "messageHistory": messageHistory,
                        "intermediateSteps": intermediateSteps,
                    },
                    "evallist": evals,
                    "sessionID": self.sessionID
                },
                "type": "FULL_EVAL_RUN",
            }

            # Add to batch
            currentBatch.append(templateData)
            if len(currentBatch) == 10:
                self.allBatches.append(currentBatch)
                currentBatch = []

        if currentBatch:
            self.allBatches.append(currentBatch)

        for batch in tqdm(
            self.allBatches,
            desc="Processing Batches",
            unit="batch",
            colour="magenta",
            ascii=False,
        ):
            try:
                self.postBatch(batch=batch, workspaceID=workspaceID)
                time.sleep(2)
                # print(batch)
            except Exception as e:
                print(f"Error posting batch: {e}")
                raise

        # Wait for results
        time.sleep(3)
        listener_thread.join()

        rawResults = self.socket.getReceivedData()

        print(f"Total results received: {len(rawResults)}")
        # print("Raw results:", rawResults)

        # print("data from db #####################",dataFromDb)
        # Fix here: keep full keys, do not split keys
        receivedRowIDs = {key for item in rawResults for key in item.keys()}
        print("Received Row IDs:", receivedRowIDs)
        expectedRowIDs = set(rowIdMapping.keys())
        missingRowIDs = expectedRowIDs - receivedRowIDs
        # print("All expected keys:", expectedRowIDs)
        # print("All received keys:", receivedRowIDs)
        # print("Missing keys:", len(missingRowIDs))
        missingRowIDs = list(missingRowIDs)

        print("Missing Row IDs:", missingRowIDs)
        print(f"Total results before fetching missing data: {len(rawResults)}")
        if len(missingRowIDs) > 0:
            print('''It's taking longer than expected to get results for some rows. You can close this now. 
                  Please wait for 15 mins while we create the flow graph for you. You can check the graph at app.llumo.ai/debugger''')
        else:
            print('''All results received successfully. You can check flowgraph in 5 mins at app.llumo.ai/debugger''')
        # if len(missingRowIDs) > 0:
        #     dataFromDb = self.fetchDataForMissingKeys(workspaceID, missingRowIDs)
        #     # print("Fetched missing data from DB:", dataFromDb)
        #     rawResults.extend(dataFromDb)
        #     print(f"Total results after fetching missing data: {len(rawResults)}")
        
        self.evalData = rawResults
        # print("RAW RESULTS: ", self.evalData)

        # Initialize dataframe columns for each eval
        for ev_name in evals:
            dataframe[ev_name] = ""
            dataframe[f"{ev_name} Reason"] = ""
            # dataframe[f"{ev_name} EdgeCase"] = None

        # Map results to dataframe rows
        for item in rawResults:
            for compound_key, value in item.items():
                if compound_key not in rowIdMapping:
                    continue
                index = rowIdMapping[compound_key]["index"]
                rowID, columnID, _ = compound_key.split("-", 2)

                # get the dataframe row at this index
                row = dataframe.iloc[index].to_dict()

                if not value:
                    continue


                # ️ Handle fullEval block
                fullEval = value.get("fullEval") if isinstance(value, dict) else None
                if fullEval:
                    if "evalMetrics" in fullEval and isinstance(fullEval["evalMetrics"], list):
                        for evalItem in fullEval["evalMetrics"]:
                            evalName = evalItem.get("evalName") or evalItem.get("kpiName")
                            score = str(evalItem.get("score")) or evalItem.get("value")
                            reasoning = evalItem.get("reasoning")
                            # edgeCase = eval_item.get("edgeCase")

                            if evalName:
                                dataframe.at[index, evalName] = score
                                dataframe.at[index, f"{evalName} Reason"] = reasoning
                                # dataframe.at[index, f"{evalName} EdgeCase"] = edgeCase

                                
                # runLog = value.get("runLog") if isinstance(value, dict) else None
                # if runLog:
                #     try:
                #         self.createRunForEvalMultiple(smartLog=runLog)
                #     except Exception as e:
                #         print(f"Error posting smartlog: {e}")
                        

                
        try:
            self.socket.disconnect()
        except Exception:
            pass

        # if hasattr(self, "endLlumoRun"):
        #     self.endEvalRun()
        #
        return dataframe

    def promptSweep(
        self,
        templates: List[str],
        data,
        model_aliases: List[AVAILABLEMODELS],
        evals=["Response Correctness"],
        toEvaluate: bool = False,
        createExperiment: bool = False,
        getDataFrame=False,
    ) -> pd.DataFrame:
        if isinstance(data, dict):
            data = [data]
        # Check if data is now a list of dictionaries
        if isinstance(data, list) and all(isinstance(item, dict) for item in data):
            working_df = pd.DataFrame(data).astype(str)
        else:
            raise ValueError("Data must be a dictionary or a list of dictionaries.")
        modelStatus = validateModels(model_aliases=model_aliases)
        if modelStatus["status"] == False:
            raise LlumoAIError.providerError(modelStatus["message"])

        self.validateApiKey()
        workspaceID = self.workspaceID
        email = self.email
        executor = ModelExecutor(self.apiKey)
        prompt_template = templates[0]

        working_df = self._outputForStream(working_df, model_aliases, prompt_template)

        # Optional evaluation
        outputEvalMapping = None
        if toEvaluate:
            for evalName in evals:
                # Validate API and dependencies
                self.validateApiKey(evalName=evalName)
                metricDependencies = checkDependency(
                    evalName, list(working_df.columns), tocheck=False
                )
                if not metricDependencies["status"]:
                    raise LlumoAIError.dependencyError(metricDependencies["message"])

            working_df, outputEvalMapping = self._evaluateForStream(
                working_df, evals, model_aliases, prompt_template, generateOutput=True
            )
        if createExperiment:
            # df = working_df.fillna("Some error occured").astype(object)
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                df = working_df.fillna("Some error occurred").astype(str)
            if createPlayground(
                email,
                workspaceID,
                df,
                promptText=prompt_template,
                definationMapping=self.definationMapping,
                evalOutputMap=outputEvalMapping,
            ):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )

        else:
            if getDataFrame == True and toEvaluate == True:
                return LlumoDataFrameResults(
                    working_df,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )

            elif getDataFrame == False and toEvaluate == True:
                data = working_df.to_dict(orient="records")
                return LlumoDictResults(
                    data,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )

            elif getDataFrame == True and toEvaluate == False:
                return working_df

            elif getDataFrame == False and toEvaluate == False:
                return working_df.to_dict(orient="records")

    # this function generates an output using llm and tools and evaluate that output
    def evaluateAgents(
        self,
        data,
        model,
        agents,
        model_api_key=None,
        evals=["Final Task Alignment"],
        prompt_template="Give answer for the given query: {{query}}",
        createExperiment: bool = False,
        getDataFrame: bool = False,
    ):
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        if model.lower() not in ["openai", "google"]:
            raise ValueError("Model must be 'openai' or 'google'")

        # converting into pandas dataframe object
        dataframe = pd.DataFrame(data)

        # Run unified agent execution
        toolResponseDf = LlumoAgentExecutor.run(
            dataframe, agents, model=model, model_api_key=model_api_key
        )

        # for eval in evals:
        # Perform evaluation
        # toolResponseDf = self.evaluate(
        #     toolResponseDf.to_dict(orient = "records"),
        #     eval=eval,
        #     prompt_template=prompt_template,
        #     createExperiment=False,
        # )
        toolResponseDf = self.evaluateMultiple(
            toolResponseDf.to_dict(orient="records"),
            evals=evals,
            prompt_template=prompt_template,
            createExperiment=createExperiment,
            getDataFrame=getDataFrame,
        )

        return toolResponseDf
        # if createExperiment:
        #     pd.set_option("future.no_silent_downcasting", True)
        #     df = toolResponseDf.fillna("Some error occured")
        #     if createPlayground(self.email, self.workspaceID, df,promptText=prompt_template,definationMapping=self.definationMapping):
        #         print(
        #             "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
        #         )
        # else:
        #     return toolResponseDf

    # this function evaluate that tools output given by the user
    def evaluateAgentResponses(
        self,
        data,
        evals=["Final Task Alignment"],
        createExperiment: bool = False,
        getDataFrame=False,
        outputColName="output",
    ):
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        dataframe = pd.DataFrame(data)

        try:
            if "query" and "messageHistory" and "tools" not in dataframe.columns:
                raise ValueError(
                    "DataFrame must contain 'query', 'messageHistory','output' ,and 'tools' columns. Make sure the columns names are same as mentioned here."
                )

            toolResponseDf = dataframe.copy()
            # for eval in evals:
            #     # Perform evaluation
            #     toolResponseDf = self.evaluate(
            #         toolResponseDf.to_dict(orient = "records"), eval=eval, prompt_template="Give answer for the given query: {{query}}",outputColName=outputColName
            #     )
            toolResponseDf = self.evaluateMultiple(
                toolResponseDf.to_dict(orient="records"),
                evals=evals,
                prompt_template="Give answer for the given query: {{query}}",
                outputColName=outputColName,
                createExperiment=createExperiment,
                getDataFrame=getDataFrame,
            )
            if createExperiment:
                pass
            else:
                return toolResponseDf

        except Exception as e:
            raise e

    def ragSweep(
        self,
        data,
        streamName: str,
        queryColName: str = "query",
        createExperiment: bool = False,
        modelAliases=[],
        prompt_template="Give answer to the given: {{query}} using the context:{{context}}",
        evals=["Context Utilization"],
        toEvaluate=False,
        generateOutput=True,
        getDataFrame=False,
    ):
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        # Validate required parameters
        if generateOutput:
            if not modelAliases:
                raise ValueError(
                    "Model aliases must be provided when generateOutput is True."
                )
            if (
                not self.apiKey
                or not isinstance(self.apiKey, str)
                or self.apiKey.strip() == ""
            ):
                raise ValueError(
                    "Valid API key must be provided when generateOutput is True."
                )

        modelStatus = validateModels(model_aliases=modelAliases)
        if modelStatus["status"] == False:
            if len(modelAliases) == 0:
                raise LlumoAIError.providerError("No model selected.")
            else:
                raise LlumoAIError.providerError(modelStatus["message"])

        # Copy the original dataframe
        original_df = pd.DataFrame(data)
        working_df = original_df.copy()

        # Connect to socket
        self.socket = LlumoSocketClient(socketUrl)
        try:
            socketID = self.socket.connect(timeout=150)
        except Exception as e:
            socketID = "DummySocketID"
        # waited_secs = 0
        # while not self.socket._connection_established.is_set():
        #     time.sleep(0.1)
        #     waited_secs += 0.1
        #     if waited_secs >= 20:
        #         raise RuntimeError("Timeout waiting for server 'connection-established' event.")

        self.validateApiKey()

        # Check user credits
        userHits = checkUserHits(
            self.workspaceID,
            self.hasSubscribed,
            self.trialEndDate,
            self.subscriptionEndDate,
            self.hitsAvailable,
            len(working_df),
        )
        if not userHits["success"]:
            raise LlumoAIError.InsufficientCredits(userHits["message"])

        print("====🚀Sit back while we fetch data from the stream 🚀====")
        workspaceID, email = self.workspaceID, self.email
        activePlayground = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
        streamId = getStreamId(workspaceID, self.apiKey, streamName)

        # Prepare batches
        rowIdMapping = {}
        self.allBatches = []
        currentBatch = []

        expectedResults = len(working_df)
        timeout = max(100, min(150, expectedResults * 10))

        listener_thread = threading.Thread(
            target=self.socket.listenForResults,
            kwargs={
                "min_wait": 40,
                "max_wait": timeout,
                "inactivity_timeout": 10,
                "expected_results": expectedResults,
            },
            daemon=True,
        )
        listener_thread.start()

        for index, row in working_df.iterrows():
            rowID, columnID = uuid.uuid4().hex, uuid.uuid4().hex
            compoundKey = f"{rowID}-{columnID}-{columnID}"
            rowIdMapping[compoundKey] = {"index": index}
            templateData = {
                "processID": getProcessID(),
                "socketID": socketID,
                "processData": {
                    "executionDependency": {"query": row[queryColName]},
                    "dataStreamID": streamId,
                },
                "workspaceID": workspaceID,
                "email": email,
                "type": "DATA_STREAM",
                "playgroundID": activePlayground,
                "processType": "DATA_STREAM",
                "rowID": rowID,
                "columnID": columnID,
                "source": "SDK",
            }
            currentBatch.append(templateData)
            if len(currentBatch) == 10 or index == len(working_df) - 1:
                self.allBatches.append(currentBatch)
                currentBatch = []

        for batch in tqdm(
            self.allBatches,
            desc="Processing Batches",
            unit="batch",
            colour="magenta",
            ncols=80,
        ):
            try:
                self.postDataStream(batch=batch, workspaceID=workspaceID)
                time.sleep(3)
            except Exception as e:
                print(f"Error posting batch: {e}")
                raise

        time.sleep(3)
        listener_thread.join()

        rawResults = self.socket.getReceivedData()
        expectedRowIDs = set(rowIdMapping.keys())
        receivedRowIDs = {key for item in rawResults for key in item.keys()}
        missingRowIDs = list(expectedRowIDs - receivedRowIDs)

        if missingRowIDs:
            dataFromDb = fetchData(workspaceID, activePlayground, missingRowIDs)
            rawResults.extend(dataFromDb)

        working_df["context"] = None
        for item in rawResults:
            for compound_key, value in item.items():
                if compound_key in rowIdMapping:
                    idx = rowIdMapping[compound_key]["index"]
                    working_df.at[idx, "context"] = value.get("value")

        # Output generation
        if generateOutput == True:
            working_df = self._outputForStream(
                working_df, modelAliases, prompt_template
            )

        # Optional evaluation
        outputEvalMapping = None
        if toEvaluate:
            for evalName in evals:
                # Validate API and dependencies
                self.validateApiKey(evalName=evalName)
                metricDependencies = checkDependency(
                    evalName, list(working_df.columns), tocheck=False
                )
                if not metricDependencies["status"]:
                    raise LlumoAIError.dependencyError(metricDependencies["message"])

            working_df, outputEvalMapping = self._evaluateForStream(
                working_df, evals, modelAliases, prompt_template, generateOutput
            )

        self.socket.disconnect()
        # Create experiment if required
        if createExperiment:
            # df = working_df.fillna("Some error occured").astype(object)
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                df = working_df.fillna("Some error occurred").astype(str)
            if createPlayground(
                email,
                workspaceID,
                df,
                queryColName=queryColName,
                dataStreamName=streamId,
                promptText=prompt_template,
                definationMapping=self.definationMapping,
                evalOutputMap=outputEvalMapping,
            ):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )
                if getDataFrame == True and toEvaluate == True:
                    return LlumoDataFrameResults(
                        working_df,
                        evals=self.evals,
                        evalData=self.evalData,
                        definationMapping=self.definationMapping,
                    )

                elif getDataFrame == False and toEvaluate == True:
                    data = working_df.to_dict(orient="records")
                    return LlumoDictResults(
                        data,
                        evals=self.evals,
                        evalData=self.evalData,
                        definationMapping=self.definationMapping,
                    )

                elif getDataFrame == True and toEvaluate == False:
                    return working_df

                elif getDataFrame == False and toEvaluate == False:
                    return working_df.to_dict(orient="records")
        else:
            if getDataFrame == True and toEvaluate == True:
                return LlumoDataFrameResults(
                    working_df,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )

            elif getDataFrame == False and toEvaluate == True:
                data = working_df.to_dict(orient="records")
                return LlumoDictResults(
                    data,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )

            elif getDataFrame == True and toEvaluate == False:
                return working_df

            elif getDataFrame == False and toEvaluate == False:
                return working_df.to_dict(orient="records")

    def _outputForStream(self, df, modelAliases, prompt_template):
        executor = ModelExecutor(self.apiKey)

        for indx, row in df.iterrows():
            inputVariables = re.findall(r"{{(.*?)}}", prompt_template)
            if not all([k in df.columns for k in inputVariables]):
                raise LlumoAIError.InvalidPromptTemplate()

            inputDict = {key: row[key] for key in inputVariables}
            for i, model in enumerate(modelAliases, 1):
                try:

                    provider = getProviderFromModel(model)
                    if provider == Provider.OPENAI:
                        validateOpenaiKey(self.apiKey)
                    elif provider == Provider.GOOGLE:
                        validateGoogleKey(self.apiKey)

                    filled_template = getInputPopulatedPrompt(
                        prompt_template, inputDict
                    )
                    response = executor.execute(provider, model.value, filled_template)
                    df.at[indx, f"output_{i}"] = response

                except Exception as e:
                    # df.at[indx, f"output_{i}"] = str(e)
                    raise e

        return df

    def _evaluateForStream(
        self, df, evals, modelAliases, prompt_template, generateOutput
    ):
        dfWithEvals = df.copy()
        outputColMapping = {}

        if generateOutput:
            # Evaluate per model output
            for i, model in enumerate(modelAliases, 1):
                outputColName = f"output_{i}"
                try:
                    res = self.evaluateMultiple(
                        dfWithEvals.to_dict("records"),
                        evals=evals,
                        prompt_template=prompt_template,
                        outputColName=outputColName,
                        _tocheck=False,
                        getDataFrame=True,
                        createExperiment=False,
                    )

                    for evalMetric in evals:
                        scoreCol = f"{evalMetric}"
                        reasonCol = f"{evalMetric} Reason"
                        if scoreCol in res.columns:
                            res = res.rename(columns={scoreCol: f"{scoreCol}_{i}"})
                        if reasonCol in res.columns:
                            res = res.rename(
                                columns={reasonCol: f"{evalMetric}_{i} Reason"}
                            )

                        outputColMapping[f"{scoreCol}_{i}"] = outputColName

                    newCols = [
                        col for col in res.columns if col not in dfWithEvals.columns
                    ]
                    dfWithEvals = pd.concat([dfWithEvals, res[newCols]], axis=1)

                except Exception as e:
                    print(f"Evaluation failed for model {model.value}: {str(e)}")

        else:
            # Evaluate only once on "output" column
            try:
                outputColName = "output"
                res = self.evaluateMultiple(
                    dfWithEvals.to_dict("records"),
                    evals=evals,
                    prompt_template=prompt_template,
                    outputColName=outputColName,
                    _tocheck=False,
                    getDataFrame=True,
                    createExperiment=False,
                )
                for evalMetric in evals:
                    scoreCol = f"{evalMetric}"
                    reasonCol = f"{evalMetric} Reason"
                    outputColMapping[scoreCol] = "output"

                newCols = [col for col in res.columns if col not in dfWithEvals.columns]
                dfWithEvals = pd.concat([dfWithEvals, res[newCols]], axis=1)
            except Exception as e:
                print(f"Evaluation failed: {str(e)}")

        return dfWithEvals, outputColMapping

    def runDataStream(
        self,
        data,
        streamName: str,
        queryColName: str = "query",
        createExperiment: bool = False,
        getDataFrame=False,
    ):

        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")

        # Copy the original dataframe
        original_df = pd.DataFrame(data)
        working_df = original_df.copy()

        # Connect to socket
        self.socket = LlumoSocketClient(socketUrl)
        try:
            socketID = self.socket.connect(timeout=150)
        except Exception as e:
            socketID = "DummySocketID"
        # waited_secs = 0
        # while not self.socket._connection_established.is_set():
        #     time.sleep(0.1)
        #     waited_secs += 0.1
        #     if waited_secs >= 20:
        #         raise RuntimeError("Timeout waiting for server 'connection-established' event.")

        self.validateApiKey()

        # Check user credits
        userHits = checkUserHits(
            self.workspaceID,
            self.hasSubscribed,
            self.trialEndDate,
            self.subscriptionEndDate,
            self.hitsAvailable,
            len(working_df),
        )
        if not userHits["success"]:
            raise LlumoAIError.InsufficientCredits(userHits["message"])

        print("====🚀Sit back while we fetch data from the stream 🚀====")
        workspaceID, email = self.workspaceID, self.email
        activePlayground = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace("-", "")
        streamId = getStreamId(workspaceID, self.apiKey, streamName)

        # Prepare batches
        rowIdMapping = {}
        self.allBatches = []
        currentBatch = []

        expectedResults = len(working_df)
        timeout = max(100, min(150, expectedResults * 10))

        listener_thread = threading.Thread(
            target=self.socket.listenForResults,
            kwargs={
                "min_wait": 40,
                "max_wait": timeout,
                "inactivity_timeout": 10,
                "expected_results": expectedResults,
            },
            daemon=True,
        )
        listener_thread.start()

        for index, row in working_df.iterrows():
            rowID, columnID = uuid.uuid4().hex, uuid.uuid4().hex
            compoundKey = f"{rowID}-{columnID}-{columnID}"
            rowIdMapping[compoundKey] = {"index": index}
            templateData = {
                "processID": getProcessID(),
                "socketID": socketID,
                "processData": {
                    "executionDependency": {"query": row[queryColName]},
                    "dataStreamID": streamId,
                },
                "workspaceID": workspaceID,
                "email": email,
                "type": "DATA_STREAM",
                "playgroundID": activePlayground,
                "processType": "DATA_STREAM",
                "rowID": rowID,
                "columnID": columnID,
                "source": "SDK",
            }
            currentBatch.append(templateData)
            if len(currentBatch) == 10 or index == len(working_df) - 1:
                self.allBatches.append(currentBatch)
                currentBatch = []

        for batch in tqdm(
            self.allBatches,
            desc="Processing Batches",
            unit="batch",
            colour="magenta",
            ncols=80,
        ):
            try:
                self.postDataStream(batch=batch, workspaceID=workspaceID)
                time.sleep(3)
            except Exception as e:
                print(f"Error posting batch: {e}")
                raise

        time.sleep(3)
        listener_thread.join()

        rawResults = self.socket.getReceivedData()
        expectedRowIDs = set(rowIdMapping.keys())
        receivedRowIDs = {key for item in rawResults for key in item.keys()}
        missingRowIDs = list(expectedRowIDs - receivedRowIDs)

        if missingRowIDs:
            dataFromDb = fetchData(workspaceID, activePlayground, missingRowIDs)
            rawResults.extend(dataFromDb)

        working_df["context"] = None
        for item in rawResults:
            for compound_key, value in item.items():
                if compound_key in rowIdMapping:
                    idx = rowIdMapping[compound_key]["index"]
                    working_df.at[idx, "context"] = value.get("value")

        self.socket.disconnect()

        # Create experiment if required
        if createExperiment:
            df = working_df.fillna("Some error occured").astype(object)
            if createPlayground(
                email,
                workspaceID,
                df,
                queryColName=queryColName,
                dataStreamName=streamId,
                definationMapping=self.definationMapping,
            ):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )
                if getDataFrame:
                    return working_df

                else:
                    data = working_df.to_dict(orient="records")
                    return data
        else:
            if getDataFrame:
                return working_df

            else:
                data = working_df.to_dict(orient="records")
                return data
            # self.latestDataframe = working_df
            # return working_df

    def createExperiment(self, dataframe):
        try:
            self.validateApiKey()

            flag = createPlayground(self.email, self.workspaceID, dataframe)
            if flag:
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )
        except Exception as e:
            raise "Some error ocuured please check your API key"

    def uploadfile(self, file_path):

        workspaceID = None
        email = None

        try:
            self.validateApiKey()
        except Exception as e:
            if hasattr(e, "response") and getattr(e, "response", None) is not None:
                pass
            raise

        # Get file extension
        _, ext = os.path.splitext(file_path)
        ext = ext.lower()

        # Supported formats
        try:
            if ext == ".csv":
                df = pd.read_csv(file_path)
            elif ext in [".xlsx", ".xls"]:
                df = pd.read_excel(file_path)
            elif ext == ".json":
                df = pd.read_json(file_path, orient="records")
            elif ext == ".parquet":
                df = pd.read_parquet(file_path)
            else:
                raise ValueError(f"Unsupported file format: {ext}")

            # If successfully loaded, call createPlayground
            df = df.astype(str)
            if createPlayground(self.email, self.workspaceID, df):

                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )

                return True

        except Exception as e:
            print(f"Error: {e}")

    def upload(self, data):
        try:
            if isinstance(data, dict):
                data = [data]
            # Check if data is now a list of dictionaries
            if isinstance(data, list) and all(isinstance(item, dict) for item in data):
                dataframe = pd.DataFrame(data).astype(str)
            else:
                raise ValueError("Data must be a dictionary or a list of dictionaries.")
            self.validateApiKey()
            if createPlayground(self.email, self.workspaceID, dataframe):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )
                return True

        except Exception as e:
            print(f"Error: {e}")
            return False

    def createExperimentWithEvals(
        self,
        data,
        evals: list,  # list of eval metric names
        prompt_template="Give answer to the given query: {{query}} using the given context: {{context}}.",
        outputColName="output",
        createExperiment: bool = False,
        getDataFrame: bool = False,
        _tocheck=True,
    ):
        if isinstance(data, dict):
            data = [data]
        elif not isinstance(data, list):
            raise ValueError("Data should be a dict or a list of dicts.")
        dataframe = pd.DataFrame(data).astype(str)
        workspaceID = None
        email = None
        self.evalData = []
        self.evals = evals
        self.allBatches = []
        rowIdMapping = {}  # (rowID-columnID-columnID -> (index, evalName))
        self.validateApiKey(evalName=evals[0])
        if createExperiment:
            if self.playgroundID:
                activePlayground = self.playgroundID
            else:
                activePlayground = str(
                    createEvalPlayground(email=self.email, workspaceID=self.workspaceID)
                )
        else:
            activePlayground = f"{int(time.time() * 1000)}{uuid.uuid4()}".replace(
                "-", ""
            )
        for evalName in evals:
            self.validateApiKey(evalName=evalName)
        self.evalData = dataframe.to_dict(orient="records")
        if createExperiment:
            print("heading to upload")
            pd.set_option("future.no_silent_downcasting", True)
            # df = dataframe.fillna("Some error occured").astype(object)
            with warnings.catch_warnings():
                warnings.simplefilter(action="ignore", category=FutureWarning)
                df = dataframe.fillna("Some error occurred").astype(str)

            df = dataframe.fillna("Some error occured").infer_objects(copy=False)
            if createPlayground(
                self.email,
                self.workspaceID,
                df,
                promptText=prompt_template,
                definationMapping=self.definationMapping,
                outputColName=outputColName,
                activePlayground=activePlayground,
            ):
                print(
                    "LLUMO’s intuitive UI is ready—start exploring and experimenting with your logs now. Visit https://app.llumo.ai/evallm to see the results."
                )

        else:
            if getDataFrame:
                return LlumoDataFrameResults(
                    dataframe,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )
            else:
                data = dataframe.to_dict(orient="records")
                return LlumoDictResults(
                    data,
                    evals=self.evals,
                    evalData=self.evalData,
                    definationMapping=self.definationMapping,
                )


class SafeDict(dict):
    def __missing__(self, key):
        return ""



import contextvars
import uuid
from typing import Optional, List, Dict, Any
from datetime import datetime, timezone
import requests
from .client import LlumoClient

_ctxLogger = contextvars.ContextVar("ctxLogger")
_ctxSessionID = contextvars.ContextVar("ctxSessionID")
_ctxLlumoRun = contextvars.ContextVar("ctxLlumoRun")


def getLogger():
    return _ctxLogger.get()


def getSessionID():
    return _ctxSessionID.get()


def getLlumoRun():
    return _ctxLlumoRun.get()


class LlumoSessionContext(LlumoClient):
    def __init__(self, logger, sessionID: Optional[str] = None):
        super().__init__(api_key=logger.apiKey, playground_id=logger.getPlaygroundID())
        self.sessionID = sessionID or str(uuid.uuid4().hex[:14])
        self.logger = logger
        self.apiKey = logger.apiKey
        self.threadLogger = None
        self.threadSessionID = None
        self.threadLlumoRun = None

    def start(self):
        self.threadLogger = _ctxLogger.set(self.logger)
        self.threadSessionID = _ctxSessionID.set(self.sessionID)

    def end(self):
        if self.threadLogger:
            _ctxLogger.reset(self.threadLogger)
        if self.threadSessionID:
            _ctxSessionID.reset(self.threadSessionID)
        if self.threadLlumoRun:
            _ctxLlumoRun.reset(self.threadLlumoRun)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, excType, excVal, excTb):
        self.end()


    def startLlumoRun(self, runName: str, rowID: str = "", columnID: str = "", runID: str = None):

        if runID is None:
            LlumoRunID = str(uuid.uuid4().hex[:16])
        else:
            LlumoRunID = runID


        # Proceed with using LlumoRunID, rowID, columnID...
        # if  rowID =="":
        #     rowID = str(uuid.uuid4().hex[:16])
        # if columnID == "":
        #     columnID = str(uuid.uuid4().hex[:16])

        currentTime = datetime(2025, 8, 2, 10, 20, 15, tzinfo=timezone.utc)
        createdAt = currentTime.strftime("%Y-%m-%dT%H:%M:%S.000Z")
        llumoRun = {
            "logID": LlumoRunID,
            "runName": runName,
            "sessionID": self.sessionID,
            "playgroundID": self.logger.getPlaygroundID(),
            "workspaceID": self.logger.getWorkspaceID(),
            "source": "SDK",
            "rowID": rowID,
            "columnID": columnID,
            "email": self.logger.getUserEmailID(),
            "createdAt": createdAt,
            "createdBy": self.logger.getUserEmailID(),
            "status": "SUCCESS",
            "flow": [],
            "latency": 4200,
            "feedback": "",
            "dump": "",
            "steps": [],
        }
        self.threadLlumoRun = _ctxLlumoRun.set(llumoRun)

    def endLlumoRun(self):
        run = getLlumoRun()
        if run is None:
            return

        # STEP 1: Sort steps by timestamp
        steps = run.get("steps", [])
        sorted_steps = sorted(steps, key=lambda s: s.get("timestamp", 0))

        # STEP 2: Remove timestamp from each step before sending
        clean_steps = [
            {k: v for k, v in step.items() if k != "timestamp"} for step in sorted_steps
        ]
        run["steps"] = clean_steps

        # print(run["runName"])  # optional debug log

        # STEP 3: Send the payload
        url = "https://app.llumo.ai/api/create-debug-log"
        headers = {
            "Authorization": f"Bearer {self.logger.getWorkspaceID()}",
            "Content-Type": "application/json",
        }

        try:
            print(run)
            response = requests.post(url, headers=headers, json=run, timeout=20)
            response.raise_for_status()
            # print(response.json())
        except requests.exceptions.Timeout:
            # print("Request timed out.")
            pass
        except requests.exceptions.RequestException as e:
            pass

        # Cleanup
        if self.threadLlumoRun:
            _ctxLlumoRun.reset(self.threadLlumoRun)
            self.threadLlumoRun = None

    def endEvalRun(self):
        run = getLlumoRun()
        if run is None:
            return

        # STEP 1: Sort steps by timestamp
        steps = run.get("steps", [])
        # sorted_steps = sorted(steps, key=lambda s: s.get("timestamp", 0))

        # # STEP 2: Remove timestamp from each step before sending
        # clean_steps = [
        #     {k: v for k, v in step.items() if k != "timestamp"} for step in sorted_steps
        # ]
        # run["steps"] = clean_steps

        # print(run["runName"])  # optional debug log

        # STEP 3: Send the payload
        url = "https://backend-api.llumo.ai/api/v1/create-debug-log-for-sdk"
        headers = {
            "Authorization": f"Bearer {self.logger.getWorkspaceID()}",
            "Content-Type": "application/json",
        }
        # print(run)
        try:
            response = requests.post(url, headers=headers, json={"log":run}, timeout=20)
            response.raise_for_status()
            # print(response.json())
        except requests.exceptions.Timeout:
            print("Request timed out.")
        except requests.exceptions.RequestException as e:
            pass

        # Cleanup
        if self.threadLlumoRun:
            _ctxLlumoRun.reset(self.threadLlumoRun)
            self.threadLlumoRun = None

    def logStep(
        self,
        stepType: str,
        stepName: str,
        metadata: Optional[dict] = None,
    ):
        # print(f"logged: {stepType}")
        run = getLlumoRun()
        if run is None:
            raise RuntimeError("No active run to log steps.")

        # add step
        stepData = {
            "stepType": stepType,
            "stepName": stepName,
            "status": metadata.get("status", "SUCCESS"),
            "message": metadata.get("message", ""),
            "metadata": metadata or {},
            "timestamp": datetime.now(timezone.utc).timestamp(),  # OPTIONAL
        }
        run["steps"].append(stepData)
        # set to context vars again in llumo run
        self.threadLlumoRun = _ctxLlumoRun.set(run)

    def logLlmStep(
        self,
        stepName: str,
        model: str,
        provider: str,
        inputTokens: int,
        outputTokens: int,
        # temperature: float,
        # promptTruncated: bool,
        latencyMs: int,
        prompt: str,
        output: str,
        status: str,
        # message: str,
    ):
        metadata = {
            "model": model,
            "provider": provider,
            "inputTokens": inputTokens,
            "outputTokens": outputTokens,
            # "temperature": temperature,
            # "promptTruncated": promptTruncated,
            "latencyMs": latencyMs,
            "prompt": prompt,
            "output": output,
            "status": status,
            # "message": message,
        }

        self.logStep("LLM", stepName, metadata)

    def logRetrieverStep(
        self,
        stepName: str,
        retrieverSource: str,
        topK: int,
        chunkSize,
        context : str,
        searchQuery: str,
        latencyMs: int,
        status: str
    ):
        metadata = {
            "retrieverSource": retrieverSource,
            "topK": topK,
            "chunkSize":chunkSize,
            "context": context,
            "searchQuery": searchQuery,
            "latencyMs": latencyMs,
            "status": status,
            # "message": message,
        }

        self.logStep("RETRIEVER", stepName, metadata)

    def logAgentStep(
        self,
        stepName: str,
        agentType: str,
        agentName: str,
        numStepsTaken: int,
        tools: List[str],
        query: str,
        status: str,
        # message: str,
    ):
        metadata = {
            "agentType": agentType,
            "agentName": agentName,
            "numStepsTaken": numStepsTaken,
            "tools": tools,
            "query": query,
            "status": status,
        #     "message": message,
         }
        self.logStep("AGENT", stepName, metadata)

    def logToolSelectorStep(
        self,
        stepName: str,
        selectorType: str,
        toolsRanked: List[Dict[str, Any]],
        selectedTool: str,
        reasoning: str,
        status: str,
        # message: str,
    ):
        metadata = {
            "selectorType": selectorType,
            "toolsRanked": toolsRanked,
            "selectedTool": selectedTool,
            "reasoning": reasoning,
            "status": status,
            # "message": message,
        }
        self.logStep("TOOL_SELECTOR", stepName, metadata)

    def logToolStep(
        self,
        stepName: str,
        toolName: str,
        description: str,
        input: Dict[str, Any],
        output: str,
        latencyMs: int,
        status: str,
        # message: str,
    ):
        metadata = {
            "toolName": toolName,
            "description":description,
            "input": input,
            "output": output,
            "latencyMs": latencyMs,
            "status": status,
            # "message": message,
        }
        self.logStep("TOOL", stepName, metadata)

    def logEvalStep(
        self,
        stepName: str,
        output: str,
        context: str,
        query: str,
        # total 7 keys add 4 more
        messageHistory: str,
        tools: str,
        intermediateSteps: str,
        groundTruth: str,
        analyticsScore: Dict[str, float],
        reasoning: Dict[str, str],
        classification: Dict[str, str],
        evalLabel: Dict[str, str],
        latencyMs: int,
        status: str,
        message: str,
    ):
        metadata = {
            "output": output,
            "context": context,
            "query": query,
            "messageHistory": messageHistory,
            "tools": tools,
            "intermediateSteps": intermediateSteps,
            "groundTruth": groundTruth,
            "analyticsScore": analyticsScore,
            "reasoning": reasoning,
            "classification": classification,
            "evalLabel": evalLabel,
            "latencyMs": latencyMs,
            "status": status,
            "message": message,
        }
        self.logStep("EVAL", stepName, metadata)

    def logFunctionCallStep(
        self,
        stepName: str,
        functionName: str,
        argsPassed: Dict[str, Any],
        output: Dict[str, Any],
        callMode: str,
        latencyMs: int,
        status: str,
        message: str,
    ):
        metadata = {
            "functionName": functionName,
            "argsPassed": argsPassed,
            "output": output,
            "callMode": callMode,
            "latencyMs": latencyMs,
            "status": status,
            "message": message,
        }
        self.logStep("FUNCTION_CALL", stepName, metadata)

    def logCompressionStep(
        self,
        stepName: str,
        prompt: str,
        promptTemplate: str,
        inputs: Dict[str, Any],
        compressedPrompt: str,
        inputToken: int,
        compressedToken: int,
        outputToken: int,
        output: str,
        compressedOutput: str,
        latencyMs: int,
        status: str,
        message: str,
    ):
        metadata = {
            "prompt": prompt,
            "promptTemplate": promptTemplate,
            "inputs": inputs,
            "compressedPrompt": compressedPrompt,
            "inputToken": inputToken,
            "compressedToken": compressedToken,
            "outputToken": outputToken,
            "output": output,
            "compressedOutput": compressedOutput,
            "latencyMs": latencyMs,
            "status": status,
            "message": message,
        }
        self.logStep("COMPRESSION", stepName, metadata)

    def logCustomScriptStep(
        self,
        stepName: str,
        inputs: Dict[str, Any],
        script: str,
        output: str,
        latencyMs: int,
        status: str,
        message: str,
    ):
        metadata = {
            "inputs": inputs,
            "script": script,
            "output": output,
            "latencyMs": latencyMs,
            "status": status,
            "message": message,
        }
        self.logStep("CUSTOM_SCRIPT", stepName, metadata)


    def logQueryStep(self,stepName,model,provider,inputTokens,query,status):
        metadata = {
            "model": model,
            "provider": provider,
            "inputTokens": inputTokens,
            "query": query,
            "status":status
        }
        self.logStep("QUERY", stepName, metadata)





import pandas as pd
from .helpingFuntions import *

class LlumoDataFrameResults(pd.DataFrame):
    _metadata=["evals","evalData","definationMapping"]

    def __init__(self, *args,evals=None,evalData=None,definationMapping=None,**kwargs):
        self.evals = evals or []
        self.evalData= evalData or []
        self.definationMapping= definationMapping or {}
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        # Needed so slicing operations return the same type
        return LlumoDataFrameResults

    def insights(self):
        
        if not self.evalData:
            print("No raw data available. Please run evaluateMultiple() first.")
            return None
        try:
            insights=[]
            reasonData,uniqueEdgecase=groupLogsByClass(self.evalData) # print(rawResults)
            
            for evalname in self.evals:
                uniqueclassesstring = ",".join(uniqueEdgecase.get(evalname, []))
                allReasons = []
                for edgeCase in reasonData[evalname]:
                    allReasons.extend(reasonData[evalname][edgeCase])
                
                evalDefinition = self.definationMapping.get(evalname, {}).get("definition", "")

                insights.append(getPlaygroundInsights(evalDefinition,uniqueclassesstring,allReasons))
            return insights
        except Exception as e:

            print("Can not genrate insights for this eval, please try again later.")


class LlumoDictResults(list):
    _metadata=["evals","evalData","definationMapping"]

    def __init__(self, *args,evals=None,evalData=None,definationMapping=None,**kwargs):
        self.evals = evals or []
        self.evalData= evalData or []
        self.definationMapping= definationMapping or {}
        super().__init__(*args, **kwargs) # This will handle list[dict]

    def insights(self):
        
        if not self.evalData:
            print("No raw data available. Please run evaluateMultiple() first.")
            return None
        try:
            insights=[]
            reasonData,uniqueEdgecase=groupLogsByClass(self.evalData) # print(rawResults)
            for evalname in self.evals:
                uniqueclassesstring = ",".join(uniqueEdgecase.get(evalname, []))
                allReasons = []
                for edgeCase in reasonData[evalname]:
                    allReasons.extend(reasonData[evalname][edgeCase])
                evalDefinition = self.definationMapping.get(evalname, {}).get("definition", "")
                insights.append(getPlaygroundInsights(evalDefinition,uniqueclassesstring,allReasons))
            return insights
        except Exception as e:
            print("Can not genrate insights for this eval, please try again later.")
    
    
for _cls in (LlumoDataFrameResults, LlumoDictResults):
    _cls.__name__ = "LlumoResults"
    _cls.__qualname__ = "LlumoResults"


class LlumoAIError(Exception):
    """Base class for all Llumo SDK-related errors."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    @staticmethod
    def InvalidApiKey():
        return LlumoAIError("The provided API key is invalid or unauthorized")

    @staticmethod
    def InvalidApiResponse():
        return LlumoAIError("Invalid or UnexpectedError response from the API")

    @staticmethod
    def RequestFailed(detail="The request to the API failed"):
        return LlumoAIError(f"Request to the API failed: {detail}")

    @staticmethod
    def InvalidJsonResponse():
        return LlumoAIError("The API response is not in valid JSON format")

    @staticmethod
    def UnexpectedError(detail="Metric"):
        return LlumoAIError(f"Can you please check if {detail} is written correctly. If you want to run {detail} please create a custom eval with same name of app.llumo.ai/evallm ")

    @staticmethod
    def EvalError(detail="Some error occured while processing"):
        return LlumoAIError(f"error: {detail}")

    @staticmethod
    def InsufficientCredits(details):
        return LlumoAIError(details)

        # return LlumoAIError("LLumo hits exhausted")

    @staticmethod
    def InvalidPromptTemplate():
        return LlumoAIError('''Make sure the prompt template fulfills the following criteria:
        1. All the variables should be inside double curly braces. Example: Give answer for the {{query}}, based on given {{context}}.
        2. The variables used in the prompt template must be present in the dataframe columns with the same name..
        ''')

    @staticmethod
    def modelHitsExhausted(details = "Your credits for the selected model exhausted."):
        return LlumoAIError(details)

    @staticmethod
    def dependencyError(details):
        return LlumoAIError(details)

    @staticmethod
    def providerError(details):
        return LlumoAIError(details)

    # @staticmethod
    # def dateNotFound():
    #     return LlumoAIError("Trial end date or subscription end date not found for the given user.")



import time
import uuid
import numpy as np
from datetime import datetime
from dateutil import parser
import requests
import json
import base64
import os
import re
import openai
import google.generativeai as genai
from collections import defaultdict


from .models import  _MODEL_METADATA,  AVAILABLEMODELS
subscriptionUrl = "https://app.llumo.ai/api/workspace/record-extra-usage"
getStreamdataUrl = "https://app.llumo.ai/api/data-stream/all"
createPlayUrl = "https://app.llumo.ai/api/New-Eval-API/create-new-eval-playground"
deletePlayUrl = "https://app.llumo.ai/api/New-Eval-API/new-upload-flow/delete-columnlist-in-playground"
uploadColList = (
    "https://app.llumo.ai/api/New-Eval-API/new-upload-flow/uploadColumnListInPlayground"
)
uploadRowList = (
    "https://app.llumo.ai/api/New-Eval-API/new-upload-flow/uploadRowsInDBPlayground"
)
createInsightUrl="https://app.llumo.ai/api/external/generate-insight-from-eval-for-sdk"

getCustomAnalyticsUrl="https://app.llumo.ai/api/workspace/get-all-analytics"

def getProcessID():
    return f"{int(time.time() * 1000)}{uuid.uuid4()}"


def getInputPopulatedPrompt(promptTemplate, tempObj):
    for key, value in tempObj.items():
        promptTemplate = promptTemplate.replace(f"{{{{{key}}}}}", value)
    return promptTemplate


def costColumnMapping(costResults, allProcess):
    # this dict will store cost column data for each row
    cost_cols = {}

    compressed_prompt = []
    compressed_prompt_output = []
    cost = []
    cost_saving = []

    for record in allProcess:
        cost_cols[record] = []
        for item in costResults:
            if list(item.keys())[0].split("-")[0] == record.split("-")[0]:
                cost_cols[record].append(list(item.values())[0])

    for ky, val in cost_cols.items():
        try:
            compressed_prompt.append(val[0])
        except IndexError:
            compressed_prompt.append("error occured")

        try:
            compressed_prompt_output.append(val[1])
        except IndexError:
            compressed_prompt_output.append("error occured")

        try:
            cost.append(val[2])
        except IndexError:
            cost.append("error occured")

        try:
            cost_saving.append(val[3])
        except IndexError:
            cost_saving.append("error occured")

    return compressed_prompt, compressed_prompt_output, cost, cost_saving


def checkUserHits(
    workspaceID,
    hasSubscribed,
    trialEndDate,
    subscriptionEndDate,
    remainingHits,
    datasetLength,
):
    # Get the current date (only the date part)
    current_date = datetime.now().date()

    # Parse trialEndDate if provided
    if trialEndDate is not None:
        try:
            trialEndDate = parser.parse(trialEndDate).date()
        except Exception:
            return {"success": False, "message": "Invalid trialEndDate format"}

    # Parse subscriptionEndDate if provided
    if subscriptionEndDate is not None:
        try:
            subscriptionEndDate = parser.parse(subscriptionEndDate).date()
        except Exception:
            return {"success": False, "message": "Invalid subscriptionEndDate format"}

    # If user is on a free trial
    if not hasSubscribed and trialEndDate is not None:
        if current_date > trialEndDate:
            return {"success": False, "message": "Trial expired. Access denied"}

        if remainingHits < datasetLength or remainingHits <= 0:
            return {"success": False, "message": "Trial Hits Exhausted"}

    else:
        if subscriptionEndDate and current_date > subscriptionEndDate:
            return {"success": False, "message": "Subscription expired. Access denied."}


        
        if remainingHits <= 0 or remainingHits < datasetLength:
            if workspaceID:
                workspaceID=str(workspaceID)
                headers = {
                    "Authorization": f"Bearer {base64.b64encode(workspaceID.encode()).decode()}",
                    "Content-Type": "application/json",
                }
                reqBody = {"unitsToSet": 1}
                responseBody = requests.post(
                    url=subscriptionUrl, json=reqBody, headers=headers
                )
                response = json.loads(responseBody.text)

                proceed = response.get("execution", "")
                # print(proceed)

                if proceed:
                    return {"success": True, "message": "Hits added and access granted."}
            else:
                return {"success": False, "message": "Workspace ID is required for subscription."}
    return {"success": True, "message": "Access granted."}


def getStreamId(workspaceID: str, token, dataStreamName):
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }
    reqBody = {"workspaceID": workspaceID}
    response = requests.post(url=getStreamdataUrl, json=reqBody, headers=headers)

    if response.status_code == 200:
        responseJson = response.json()
        data = responseJson.get("data", [])

        # Find stream by name
        matchedStream = next(
            (stream for stream in data if stream.get("name") == dataStreamName), None
        )

        if matchedStream:

            return matchedStream.get("dataStreamID")

        else:
            print(f"No stream found with name: {dataStreamName}")
            return None
    else:
        print("Error:", response.status_code, response.text)
        return None


def createEvalPlayground(email: str, workspaceID: str):
    url = createPlayUrl
    headers = {
        "Content-Type": "application/json",
    }
    payload = {
        "email": email,
        "workspaceID": workspaceID,
    }

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code == 200:
        try:
            responseJson = response.json()
            # print(responseJson)
            return responseJson.get("data", {}).get("playgroundID", None)

        except Exception as e:
            print("Failed to parse JSON:", e)
            return None
    else:
        print("Error:", response.status_code, response.text)
        return None


def deleteColumnListInPlayground(workspaceID: str, playgroundID: str):
    url = deletePlayUrl
    headers = {
        "Content-Type": "application/json",
    }
    payload = {
        "workspaceID": workspaceID,
        "playgroundID": playgroundID,
    }

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code == 200:
        try:

            return response.json()
        except Exception as e:
            print("⚠️ Failed to parse JSON:", e)
            return None
    else:
        print("❌ Error:", response.status_code, response.text)
        return None

def createColumn(workspaceID, dataframe, playgroundID, promptText=None,queryColName=None,
                 outputColName= "output",dataStreamName=None,definationMapping=None,evalOutputMap = None,customAnalytics=[]):
    if len(dataframe) > 100:
        dataframe = dataframe.head(100)
        print("⚠️ Dataframe truncated to 100 rows for upload.")

    coltemplate = {
        "workspaceID": workspaceID,
        "playgroundID": playgroundID,
        "columnListToUpload": [],
    }
    allEvals = ['Response Completeness', 'Response Bias', 'Response Harmfulness', 'Input Toxicity', 'Input Harmfulness', 'Context Utilization', 'Relevance Retention', 'Semantic Cohesion', 'Final Task Alignment', 'Tool Reliability', 'Response Correctness', 'Response Toxicity', 'Input Bias', 'Input Relevancy', 'Redundancy Reduction', 'Response Sentiment', 'Tool Selection Accuracy', 'Stepwise Progression', 'Hallucination', 'Faithfulness', 'Answer Relevancy', 'Context Precision', 'Answer Similarity', 'Harmfulness', 'Maliciousness', 'Coherence', 'Answer Correctness', 'Context Recall', 'Context Entity Recall', 'Conciseness', 'customEvalColumn', 'Groundedness', 'Memory Utilization', 'Input Relevancy (Multi-turn)','PII Check','Prompt Injection']
    try:
        allEvals.extend(list(customAnalytics.keys()))
    except Exception as e:
        pass
    evalDependencies =  checkDependency(_returnDepMapping=True,customevals=customAnalytics)
    # print(allEvals)
    # Create a mapping of column names to unique column IDs
    columnIDMapping = {}

    # Iterate over each column in the dataframe
    for indx, col in enumerate(dataframe.columns):
        # Generate a unique column ID using uuid
        columnID = str(uuid.uuid4().hex[:8])

        columnIDMapping[col] = columnID


        # if col.startswith('output') and promptText!=None:
        #     # For output columns, create the prompt template with promptText
        #     if promptText:
        #         # Extract variables from promptText and set them as dependencies
        #         dependencies = []
        #
        #         # Find variables inside {{variable}}
        #         variables = re.findall(r'{{(.*?)}}', promptText)
        #
        #         # Loop through each variable and check if it exists as a column name
        #         for var in variables:
        #             varName = var.strip()
        #             if varName in columnIDMapping:  # Check if the variable is a column name
        #                 dependencies.append(columnIDMapping[varName])  # Add its columnID
        #
        #         # Now update the template for the output column
        #
        #         template={
        #             "provider": "OPENAI",
        #             "model": "GPT_4o",
        #             "promptText": promptText,
        #             "modelOptions": {
        #                 "temperature": 0,
        #                 "frequencyPenalty": 0,
        #                 "presencePenalty": 0,
        #                 "maxToken": 8192
        #             },
        #             "toolConfig": "none",
        #             "concurrency": "",
        #             "outputType": "STRING",
        #             "isPromptSelected": True,
        #             "isSmartPromptSelected": False,
        #             "dependency": dependencies,  # Use the dependencies extracted from promptText
        #             "columnID": columnID,  # Use the generated column ID
        #             "label": col,
        #             "type": "PROMPT",
        #             "order": indx,
        #         }

        if col.startswith('context') and dataStreamName != None :
            if queryColName and dataStreamName:
                dependencies = []
                dependencies.append(columnIDMapping[queryColName])
                template = {
                    "variableType": "STRING",
                    "dependency": dependencies,
                    "dataStreamName": dataStreamName,
                    "query": columnIDMapping[queryColName],
                    "columnID": columnID,  # Use the generated column ID
                    "label": "context",
                    "type": "DATA_STREAM",
                    "order": indx}


        elif any(col.startswith(eval + "_") or col == eval for eval in allEvals) and not " Reason" in col and promptText is not None :
            if evalOutputMap != None:
                outputColName = evalOutputMap[col]
            else:
                outputColName = outputColName
            dependencies = []
            variables = re.findall(r'{{(.*?)}}', promptText)

                # Loop through each variable and check if it exists as a column name
            for var in variables:
                varName = var.strip()
                if varName in columnIDMapping:  # Check if the variable is a column name
                    dependencies.append(columnIDMapping[varName])

            dependencies.append(columnIDMapping[outputColName])  # Add the output column ID

            longDef = definationMapping.get(col.rsplit("_",1)[0], {}).get('definition', "")
            shortDef =definationMapping.get(col.rsplit("_",1)[0], {}).get('briefDefinition', "")
            enum = col.rsplit("_",1)[0].upper().replace(" ","_")

            template = {
      "analytics": [
        col.lower().replace(" ","_")
      ],
      "evaluationMetric": "ALL",
      "evaluationModel": "LLUMO_EVALLM",
      "selectPrompt": None if "output" not in columnIDMapping.keys() else columnIDMapping["output"],
      "scoreCondition": "GREATER_THAN",
      "scoreValue": "50",
      "scoreResult": "PASS",
      "llmKpi": col.rsplit("_",1)[0],
      "setRules": True,
      "type": "EVAL",
      "evalType": "LLM",
      "similarityMetric": None,
      "embeddingModel": None,
      "groundTruth": None if "groundTruth" not in columnIDMapping.keys() else columnIDMapping["groundTruth"],
      "dataStream": None,
      "context":None if "context" not in columnIDMapping.keys() else columnIDMapping["context"],
      "dependency":[columnIDMapping[col] if dep == "output" else columnIDMapping[dep] for dep in evalDependencies[col.rsplit("_", 1)[0]]],
      "query": None if "query" not in columnIDMapping.keys() else columnIDMapping["query"],
    "tools":None if "tools" not in columnIDMapping.keys() else columnIDMapping["tools"],
    "messageHistory":None if "messageHistory" not in columnIDMapping.keys() else columnIDMapping["messageHistory"],
      "hallucinationFields": {
        "query": None,
        "context": None,
        "output": None
      },
      "definition": longDef,
      "analyticsENUM": enum,
      "prompt": shortDef,
      "analyticsName": col.rsplit("_",1)[0],
      "columnID": columnID,
      "label": col,
      "order": indx
    }

        elif col.endswith(' Reason') and promptText!=None:
            continue


        else:

            template = {
                "label": col,  # Label is the column name
                "type": "VARIABLE",  # Default type for non-output columns
                "variableType": "STRING",
                "order": indx,
                "columnID": columnID,  # Use the generated column ID
            }

        # Append the template to the column list
        coltemplate["columnListToUpload"].append(template)

    # Prepare the row template structure
    rowTemplate = {
        "workspaceID": workspaceID,
        "playgroundID": playgroundID,
        "dataToUploadList": [],
        "columnList": coltemplate["columnListToUpload"],
    }

    # Populate dataToUploadList with rows from the dataframe
    for indx, row in dataframe.iterrows():
        row_dict = {}

        # For each column, we need to map the column ID to the corresponding value in the row
        
        for col in dataframe.columns:
            columnID = columnIDMapping[col]

            if any(col.startswith(eval + "_") or col == eval for eval in allEvals) and not " Reason"  in col and promptText!=None:
                
                row_dict[columnID] = {

                    "value": row[col],
                    "type": "EVAL",
                    "isValid": True,
                    "reasoning": row[col+" Reason"],
                    "edgeCase": "minorHallucinationDetailNotInContext",
                    "kpi": col

                }
            elif col.endswith(' Reason') and promptText!=None:
                continue
            else:# Get the columnID from the mapping
                row_dict[columnID] = row[col]

            # row_dict[columnID] = row[col]  # Directly map the column ID to the row value
        # Add the row index (if necessary)
        row_dict["pIndex"] = indx
        rowTemplate["dataToUploadList"].append(row_dict)

    # Return the column template, row template, and the column ID mapping
    return coltemplate, rowTemplate

def uploadColumnListInPlayground(payload):
    url = uploadColList
    headers = {
        "Content-Type": "application/json",
    }
    payload = payload

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code == 200:
        try:

            return response.json()
        except Exception as e:
            print("⚠️ Failed to parse JSON:", e)
            return None
    else:
        print("❌ Error:", response.status_code, response.text)
        return None


def uploadRowsInDBPlayground(payload):
    url = uploadRowList
    headers = {
        "Content-Type": "application/json",
    }

    payload = payload

    response = requests.post(url, json=payload, headers=headers)

    if response.status_code == 200:
        try:

            return response.json()
        except Exception as e:
            print("⚠️ Failed to parse JSON:", e)
            return None
    else:
        print("❌ Error:", response.status_code, response.text)
        return None


def createPlayground(email, workspaceID, df, promptText=None,queryColName=None,dataStreamName=None,definationMapping=None,outputColName="output",evalOutputMap = None,activePlayground=None,customAnalytics=[]):
    
    if activePlayground != None:
        playgroundId=activePlayground
    else:
        playgroundId = str(createEvalPlayground(email=email, workspaceID=workspaceID))
    payload1, payload2 = createColumn(
        workspaceID=workspaceID, dataframe=df, playgroundID=playgroundId, promptText=promptText,queryColName=queryColName,dataStreamName=dataStreamName,definationMapping=definationMapping,outputColName=outputColName,evalOutputMap=evalOutputMap,customAnalytics=customAnalytics
    )
    
 # Debugging line to check the payload2 structure
    deleteExistingRows = deleteColumnListInPlayground(
        workspaceID=workspaceID, playgroundID=playgroundId
    )
    colListUpload = uploadColumnListInPlayground(payload=payload1)
    rowListUpload = uploadRowsInDBPlayground(payload=payload2)

    if rowListUpload:
        return True




def getPlaygroundInsights(defination:str,uniqueClassesString: str, reasonList: list):
    headers = {
        
        "Content-Type": "application/json",
    }

    # Initial request to generate playground insights
    payload = {
        "uniqueClassesString": uniqueClassesString,
        "reasonList": reasonList,
        "definition": defination,
    }

    urlGenerate = createInsightUrl
    try:
        responseGenerate = requests.post(urlGenerate, json=payload, headers=headers)

        if responseGenerate.status_code == 200:
            responseJson = responseGenerate.json()
            # print(responseJson)

            # ✅ NEW: unwrap "data" if it exists, otherwise fall back to the old shape
            payloadData = responseJson.get("data", responseJson)

            # keep only the fields you care about
            filteredResponse = {
                key: payloadData[key] for key in ("analysis", "nextStep") if key in payloadData
            }

            return filteredResponse
    except Exception as e:
        print(f"Exception occurred while generating insight: {e}")
        return None
        
    else:
        print(f"Error generating insight: {responseGenerate.status_code} - {responseGenerate.text}")
        return None
def checkDependency(selectedEval:list = [], columns:list = [],tocheck=True,_returnDepMapping = False,customevals={}):
    """
    Checks if all the required input columns for the selected evaluation metric are present.

    Parameters:
    - selectedEval (str): The name of the selected evaluation metric.
    - columns (list): List of column names present in the dataset.

    Raises:
    - LlumoAIError.dependencyError: If any required column is missing.
    """
    # Define required dependencies for each evaluation metric

    metricDependencies = {
        'Response Completeness': ['context', 'query', 'output'],
        'Response Bias': ['output'],
        'Response Harmfulness': ['output'],
        'Input Toxicity': ['query'],
        'Input Harmfulness': ['query'],
        'Context Utilization': ['output', 'context'],
        'Relevance Retention': ['context', 'query'],
        'Semantic Cohesion': ['context'],
        'Final Task Alignment': ['query','output'],
        'Tool Reliability': ['intermediateSteps'],
        'Response Correctness': ['output', 'query', 'context'],
        'Response Toxicity': ['output'],
        'Input Bias': ['query'],
        'Input Relevancy': ['context', 'query'],
        'Redundancy Reduction': ['context'],
        'Response Sentiment': ['output'],
        'Tool Selection Accuracy': ['tools', 'intermediateSteps'],
        'Stepwise Progression': ['tools', 'intermediateSteps'],
        'Hallucination': ['query', 'context', 'output'],
        'Groundedness': ['groundTruth', 'output'],
        'Memory Utilization': ['context', 'messageHistory'],
        'Input Relevancy (Multi-turn)': ['context', 'query'],
        'PII Check':["query","output"],
        'Prompt Injection':["query"]
    }
    
    metricDependencies.update(customevals)
    if _returnDepMapping == True:
        return metricDependencies

    if tocheck == True:
        # Check if the selected evaluation metric is known
        if selectedEval not in metricDependencies:
            return {"status": False,"message":f"Unknown evaluation metric: {selectedEval}"}

        # Get the required columns for the selected evaluation
        columnsRequired = metricDependencies[selectedEval]

        # Check if each required column is present in the provided columns
        for requirement in columnsRequired:
            if requirement not in columns:
                return {"status":False,
                    "message":f"'{selectedEval}' requires columns: {columnsRequired}. "
                    f"Missing: '{requirement}'. Please ensure your data includes all required columns."
                    }
        return {"status":True,"message":"success"}
    else:
        return {"status":True,"message":"success"}
    

def fetchData(workspaceID, playgroundID, missingList: list):
    # Define the URL and prepare the payload
    socket_data_url = "https://redskull.llumo.ai/api/eval/get-awaited"
    payload = {
        "workspaceID": workspaceID,
        "playgroundID": playgroundID,
        "missingList": missingList
    }
    
    try:
        # Send a POST request to the API
        response = requests.post(socket_data_url, json=payload)
        
        # Check if the response is successful
        if response.status_code == 200:
            # Parse the JSON data from the response
            data = response.json().get("data", {})
            
            
            # Prepare the list of all data values in the desired format
            result_list = []
            for key, value in data.items():
                # Create a dictionary for each item in the response data
                result_list.append({
                    key: {
                        "value": value.get("value"),
                        "reasoning": value.get("reasoning"),
                        "edgeCase": value.get("edgeCase"),
                        "kpi": value.get("kpi")
                    }
                })
            
            return result_list
        else:
            print(f"Failed to fetch data. Status Code: {response.status_code}")
            return []

    except Exception as e:
        print(f"An error occurred: {e}")
        return []

def validateModels(model_aliases):

    selectedProviders = []
    for name in model_aliases:
        for alias ,(provider , modelName ) in _MODEL_METADATA.items():
            if modelName == name:
                selectedProviders.append(provider)

    if len(set(selectedProviders)) > 1:
        return {"status": False,"message":"All selected models should be of same provider."}
    else:
        return {"status": True,"message":"All selected models are of same provider."}



def validateOpenaiKey(api_key):
    try:
        client = openai.OpenAI(api_key=api_key)
        _ = client.models.list()  # Light call to list models
    except openai.AuthenticationError:
        raise ValueError("❌ Invalid OpenAI API key.")
    except Exception as e:
        raise RuntimeError(f"⚠️ Error validating OpenAI key: {e}")

def validateGoogleKey(api_key):
    try:
        genai.configure(api_key=api_key)
        _ = genai.GenerativeModel("gemini-2.0-flash-lite").generate_content("test")
    except Exception as e:
        if "PERMISSION_DENIED" in str(e) or "API key not valid" in str(e):
            raise ValueError("❌ Invalid Google API key.")
        raise RuntimeError(f"⚠️ Error validating Gemini key: {e}")

def groupLogsByClass(logs, max_logs=2):
    # Initialize the final result structures (no defaultdict)
    
    groupedLogs = {}
    uniqueEdgeCases = {}  # This will store unique edge cases for each eval_name

    # Iterate through the logs
    for log in logs:
        log_details = list(log.values())[0]  # Get the details dictionary
        eval_name = log_details.get("kpi", "unmarked")
        edge_case = log_details.get("edgeCase", "unmarked")
        reasoning = log_details.get("reasoning", "")

        if eval_name != "unmarked" and edge_case != "unmarked":
            # Ensure that the eval_name and edge_case exist in the dictionary
            if eval_name not in groupedLogs:
                groupedLogs[eval_name] = {}
                uniqueEdgeCases[eval_name] = set()  # Initialize the set for unique edge cases

            if edge_case not in groupedLogs[eval_name]:
                groupedLogs[eval_name][edge_case] = []

            # Append the reasoning to the correct place
            groupedLogs[eval_name][edge_case].append(reasoning)
            uniqueEdgeCases[eval_name].add(edge_case)  # Add the edge case to the set

    # Limit the number of reasons to max_logs
    for eval_name in groupedLogs:
        for edge_case in groupedLogs[eval_name]:
            groupedLogs[eval_name][edge_case] = groupedLogs[eval_name][edge_case][:max_logs]

    # Convert the set of unique edge cases to a list for easier reading
    for eval_name in uniqueEdgeCases:
        uniqueEdgeCases[eval_name] = list(uniqueEdgeCases[eval_name])

    return groupedLogs, uniqueEdgeCases


def getCustomAnalytics(workspaceID):
    try:
        url = getCustomAnalyticsUrl
        payload = {
            "workspaceID": workspaceID
        }

        headers = {
            "Content-Type": "application/json"
        }

        response = requests.post(url, json=payload, headers=headers)
        data=response.json()
        customData=data.get("data","").get("analyticsCustom","")
        customMapping = {
            "QUERY": "query",
            "CONTEXT": "context",
            "OUTPUT": "output",
            "MESSAGEHISTORY": "messageHistory",
            "TOOLS": "tools",
            "INTERMEDIATESTEPS": "intermediateSteps",
            "GROUNDTRUTH": "groundTruth",
        }

        metricDependencies = {}
        
        

        for eval in customData:
            evalName = eval.get("analyticsName")
            evalDependencyRaw = list(eval.get("variableMappings").values())
            
            # Replace each value using the custom mapping
            evalDependency = [customMapping.get(val.upper(), val.lower()) for val in evalDependencyRaw]
            
            # Build the dict
            metricDependencies[evalName] = evalDependency
        return metricDependencies
    
    except Exception as e:
        return {}


import time
from openai import OpenAI as OpenAIClient
from .client import LlumoClient
from .llumoSessionContext import LlumoSessionContext
from .llumoLogger import LlumoLogger


# evaluation function that uses LlumoClient
def performEvaluation(data, api_key=None, evals=["Response Correctness"], **kwargs):
    try:
        client = LlumoClient(api_key=api_key)
        results = client.evaluateMultiple(
            data,
            evals=evals,
            prompt_template="Give answer to the query: {{query}}, using context: {{context}}",
            getDataFrame=False,
        )
        return results
    except Exception as e:
        print(f"Error in perform_evaluation: {e}")
        raise


# Wrapper around ChatCompletion to allow custom fields like `.evaluation`
class ChatCompletionWithEval:
    def __init__(
        self,
        response,
        #  , evaluation
    ):
        self._response = response
        # self.evaluation = evaluation

    def __getattr__(self, name):
        return getattr(self._response, name)

    def __getitem__(self, key):
        return self._response[key]

    def __repr__(self):
        return repr(self._response)


class OpenAI(OpenAIClient):
    def __init__(self, api_key: str, session):
        super().__init__(api_key=api_key)
        self.session = session
        self.llumo_key = session.apiKey

        original_create = self.chat.completions.create

        def create_wrapper(*args, **kwargs):
            context = kwargs.pop("context", None)
            model = kwargs["model"]
            create_experiment = kwargs.pop("createExperiment", False)

            messages = kwargs.get("messages", [])
            user_message = next(
                (
                    m.get("content")
                    for m in reversed(messages)
                    if m.get("role") == "user"
                ),
                "",
            )

            if not context or context.strip() == "":
                context = user_message

            # Get IDs from the session logger
            playground_id = self.session.logger.getPlaygroundID()
            workspace_id = self.session.logger.getWorkspaceID()

            # Input Bias Evaluation
            # eval_input_bias = [
            #     {
            #         "query": user_message,
            #         "context": context,
            #         "output": "",  # No output yet
            #     }
            # ]
            # try:
            #     start_time = time.time()
            #     bias_evaluation_result = performEvaluation(
            #         eval_input_bias,
            #         api_key=self.llumo_key,
            #         evals=["Input Bias"],
            #         playgroundID=playground_id,
            #         workspaceID=workspace_id,
            #         createExperiment=create_experiment,
            #     )
            #     latency = int((time.time() - start_time) * 1000)
            #     # Access the first result object
            #     bias_evaluation = bias_evaluation_result[0]
            #     message = "-".join(
            #         getattr(bias_evaluation, "edgeCases", {}).get("value", [])
            #     )
            #     self.session.logEvalStep(
            #         stepName=f"EVAL-Input Bias",
            #         output="",
            #         context=context,
            #         query=user_message,
            #         messageHistory="",
            #         tools="",
            #         intermediateSteps="",
            #         groundTruth="",
            #         analyticsScore=getattr(bias_evaluation, "analyticsScore", {}),
            #         reasoning=getattr(bias_evaluation, "reasoning", {}),
            #         classification=getattr(bias_evaluation, "classification", {}),
            #         evalLabel=getattr(bias_evaluation, "evalLabel", {}),
            #         latencyMs=latency,
            #         status="SUCCESS",
            #         message=message,
            #     )
            # except Exception as e:
            #     print(f"Input Bias evaluation failed: {e}")
            #     self.session.logEvalStep(
            #         stepName=f"EVAL-FAILURE",
            #         output="",
            #         context=context,
            #         query=user_message,
            #         messageHistory="",
            #         tools="",
            #         intermediateSteps="",
            #         groundTruth="",
            #         analyticsScore={},
            #         reasoning={},
            #         classification={},
            #         evalLabel={},
            #         latencyMs=0,
            #         status="FAILURE",
            #         message="EVAL_ERROR",
            #     )

            start_time = time.time()
            response = original_create(*args, **kwargs)
            latency = int((time.time() - start_time) * 1000)
            output_text = response.choices[0].message.content
            self.session.logQueryStep(
                stepName="Query Invocation",
                model=model,
                provider="openai",
                inputTokens=response.usage.prompt_tokens,
                query=user_message,
                status = "SUCCESS")

            self.session.logLlmStep(
                stepName=f"LLM-{user_message[:30]}",
                model=model,
                provider="openai",
                inputTokens=response.usage.prompt_tokens,
                outputTokens=response.usage.completion_tokens,
                # temperature=kwargs.get("temperature", 0.0),
                # promptTruncated=False,
                latencyMs=latency,
                prompt=user_message,
                output=output_text,
                status="SUCCESS",
                # message="",
            )

            # Response Correctness Evaluation
            # eval_input_correctness = [
            #     {
            #         "query": user_message,
            #         "context": context,
            #         "output": output_text,
            #     }
            # ]
            # try:
            #     start_time = time.time()
            #     correctness_evaluation_result = performEvaluation(
            #         eval_input_correctness,
            #         api_key=self.llumo_key,
            #         evals=["Response Correctness"],
            #         playgroundID=playground_id,
            #         workspaceID=workspace_id,
            #         createExperiment=create_experiment,
            #     )
            #     latency = int((time.time() - start_time) * 1000)
            #     # Access the first result object
            #     correctness_evaluation = correctness_evaluation_result[0]
            #     message = "-".join(
            #         getattr(correctness_evaluation, "edgeCases", {}).get("value", [])
            #     )
            #     self.session.logEvalStep(
            #         stepName=f"EVAL-Response Correctness",
            #         output=output_text,
            #         context=context,
            #         query=user_message,
            #         messageHistory="",
            #         tools="",
            #         intermediateSteps="",
            #         groundTruth="",
            #         analyticsScore=getattr(
            #             correctness_evaluation, "analyticsScore", {}
            #         ),
            #         reasoning=getattr(correctness_evaluation, "reasoning", {}),
            #         classification=getattr(
            #             correctness_evaluation, "classification", {}
            #         ),
            #         evalLabel=getattr(correctness_evaluation, "evalLabel", {}),
            #         latencyMs=latency,
            #         status="SUCCESS",
            #         message=message,
            #     )
            # except Exception as e:
            #     print(f"Response Correctness evaluation failed: {e}")
            #     correctness_evaluation = None
            #     self.session.logEvalStep(
            #         stepName=f"EVAL-FAILURE",
            #         output=output_text,
            #         context=context,
            #         query=user_message,
            #         messageHistory="",
            #         tools="",
            #         intermediateSteps="",
            #         groundTruth="",
            #         analyticsScore={},
            #         reasoning={},
            #         classification={},
            #         evalLabel={},
            #         latencyMs=0,
            #         status="FAILURE",
            #         message="EVAL_ERROR",
            #     )

            # if correctness_evaluation is None:
            #     return response

            return ChatCompletionWithEval(
                response
                # , correctness_evaluation
            )

        self.chat.completions.create = create_wrapper



from langchain_core.callbacks.base import BaseCallbackHandler
from typing import Any, Dict, List, Optional, Union
import time
import json
from llumo.llumoSessionContext import  getSessionID, getLlumoRun
from llumo.llumoSessionContext import LlumoSessionContext


class LlumoCallbackHandler(BaseCallbackHandler):
    """
    LangChain callback handler that integrates with Llumo logging system.
    Tracks LLM calls, tool usage, agent actions, and chains.
    """

    def __init__(self, logger):
        self.logger = logger
        self.start_times = {}  # Track start times for latency calculation
        self.step_counters = {}  # Track step counts for agents

    def _get_session_context(self) -> Optional[LlumoSessionContext]:
        """Get the current Llumo session context from context variables."""
        try:
            session_id = getSessionID()
            run = getLlumoRun()
            if session_id and run:
                # Create a temporary context object to access logging methods
                ctx = LlumoSessionContext(self.logger, session_id)
                return ctx
        except Exception:
            pass
        return None

    def _safe_serialize(self, obj: Any) -> str:
        """Safely serialize objects to JSON string."""
        try:
            return json.dumps(obj, default=str, ensure_ascii=False)
        except Exception:
            return str(obj)

    # LLM Callbacks
    def on_llm_start(
            self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        """Called when LLM starts generating."""
        run_id = kwargs.get('run_id')
        if run_id:
            self.start_times[run_id] = time.time()

        print("LLM started - prompts:", len(prompts))

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Called when LLM generates a new token."""
        # Optional: Could be used for streaming token tracking
        pass

    def on_llm_end(self, response: Any, **kwargs: Any) -> None:
        """Called when LLM finishes generating."""
        ctx = self._get_session_context()
        if not ctx:
            print("No Llumo session context available")
            return

        run_id = kwargs.get('run_id')
        start_time = self.start_times.pop(run_id, time.time())
        latency_ms = int((time.time() - start_time) * 1000)

        # Extract LLM response details
        model = getattr(response, 'model_name', 'unknown')

        # Get token usage if available
        token_usage = getattr(response, 'llm_output', {}).get('token_usage', {})
        input_tokens = token_usage.get('prompt_tokens', 0)
        output_tokens = token_usage.get('completion_tokens', 0)

        # Get the generated text
        if hasattr(response, 'generations') and response.generations:
            output_text = response.generations[0][0].text if response.generations[0] else ""
        else:
            output_text = str(response)

        # Get the original prompt
        prompts = kwargs.get('prompts', [''])
        query = prompts[0] if prompts else ""

        try:
            ctx.logLlmStep(
                stepName=f"LLM Call - {model}",
                model=model,
                provider="langchain",
                inputTokens=input_tokens,
                outputTokens=output_tokens,
                temperature=kwargs.get('temperature', 0.7),
                promptTruncated=False,
                latencyMs=latency_ms,
                query=query,
                output=output_text,
                status="SUCCESS",
                message=""
            )
            print(f"Logged LLM step: {model} ({latency_ms}ms)")
        except Exception as e:
            print(f"Failed to log LLM step: {e}")

    def on_llm_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when LLM encounters an error."""
        ctx = self._get_session_context()
        if not ctx:
            return

        run_id = kwargs.get('run_id')
        start_time = self.start_times.pop(run_id, time.time())
        latency_ms = int((time.time() - start_time) * 1000)

        prompts = kwargs.get('prompts', [''])
        query = prompts[0] if prompts else ""

        try:
            ctx.logLlmStep(
                stepName="LLM Call - Error",
                model="unknown",
                provider="langchain",
                inputTokens=0,
                outputTokens=0,
                temperature=0.7,
                promptTruncated=False,
                latencyMs=latency_ms,
                query=query,
                output="",
                status="FAILURE",
                message=str(error)
            )
            print(f"Logged LLM error: {error}")
        except Exception as e:
            print(f"Failed to log LLM error: {e}")

    # Chain Callbacks
    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        pass

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Called when a chain ends."""
        print("Chain execution completed")

    def on_chain_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when a chain encounters an error."""
        print(f"Chain error: {error}")

    # Tool Callbacks
    def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        """Called when a tool starts."""
        run_id = kwargs.get('run_id')
        if run_id:
            self.start_times[run_id] = time.time()

        tool_name = serialized.get('name', 'Unknown Tool')
        print(f"Tool started: {tool_name}")

    def on_tool_end(self, output: str, **kwargs: Any) -> None:
        """Called when a tool ends."""
        ctx = self._get_session_context()
        if not ctx:
            return

        run_id = kwargs.get('run_id')
        start_time = self.start_times.pop(run_id, time.time())
        latency_ms = int((time.time() - start_time) * 1000)

        # Extract tool info from kwargs
        serialized = kwargs.get('serialized', {})
        tool_name = serialized.get('name', 'Unknown Tool')
        input_str = kwargs.get('input_str', '')

        try:
            ctx.logToolStep(
                stepName=f"Tool - {tool_name}",
                toolName=tool_name,
                input={"input": input_str},
                output=output,
                latencyMs=latency_ms,
                status="SUCCESS",
                message=""
            )
            print(f"Logged tool step: {tool_name} ({latency_ms}ms)")
        except Exception as e:
            print(f"Failed to log tool step: {e}")

    def on_tool_error(self, error: Exception, **kwargs: Any) -> None:
        """Called when a tool encounters an error."""
        ctx = self._get_session_context()
        if not ctx:
            return

        run_id = kwargs.get('run_id')
        start_time = self.start_times.pop(run_id, time.time())
        latency_ms = int((time.time() - start_time) * 1000)

        serialized = kwargs.get('serialized', {})
        tool_name = serialized.get('name', 'Unknown Tool')
        input_str = kwargs.get('input_str', '')

        try:
            ctx.logToolStep(
                stepName=f"Tool - {tool_name} - Error",
                toolName=tool_name,
                input={"input": input_str},
                output="",
                latencyMs=latency_ms,
                status="FAILURE",
                message=str(error)
            )
            print(f"Logged tool error: {tool_name} - {error}")
        except Exception as e:
            print(f"Failed to log tool error: {e}")

    # Agent Callbacks
    def on_agent_action(self, action: Any, **kwargs: Any) -> None:
        """Called when an agent takes an action."""
        run_id = kwargs.get('run_id')

        # Track agent step count
        if run_id not in self.step_counters:
            self.step_counters[run_id] = 0
        self.step_counters[run_id] += 1

        print(f"Agent action: {getattr(action, 'tool', 'unknown')}")

    def on_agent_finish(self, finish: Any, **kwargs: Any) -> None:
        """Called when an agent finishes."""
        ctx = self._get_session_context()
        if not ctx:
            return

        run_id = kwargs.get('run_id')
        num_steps = self.step_counters.pop(run_id, 0)

        try:
            ctx.logAgentStep(
                stepName="Agent Execution",
                agentType="langchain_agent",
                agentName="LangChain Agent",
                numStepsTaken=num_steps,
                tools=[],  # Could be populated if tool info is available
                query=getattr(finish, 'log', ''),
                status="SUCCESS",
                message=""
            )
            print(f"Logged agent finish: {num_steps} steps")
        except Exception as e:
            print(f"Failed to log agent step: {e}")









from .exceptions import LlumoAIError
from .helpingFuntions import *
from .models import AVAILABLEMODELS
from .execution import ModelExecutor
from .functionCalling import *
from .openai import OpenAI
from .google import genai
from .llumoSessionContext import *
from .llumoLogger import *
from .callback import *